from inspect import getfullargspec as fargs
from typing import List, Union, Callable

import numpy
import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import TimeDistributed

from angorapy.tasks.wrappers import TaskWrapper
from angorapy.utilities.error import IncompatibleModelException
from angorapy.utilities.core import flatten, env_extract_dims


def is_recurrent_model(model: tf.keras.Model):
    """Check if given model is recurrent (i.e. contains a recurrent layer of any sort)"""
    for layer in model.submodules:
        if isinstance(layer, tf.keras.layers.RNN):
            return True

    return False


def requires_batch_size(model_builder) -> bool:
    """Check if model building function requires a batch size when building."""
    return "bs" in fargs(model_builder).args + fargs(model_builder).kwonlyargs


def requires_sequence_length(model_builder) -> bool:
    """Check if model building function requires a sequenc_length when building."""
    return "sequence_length" in fargs(model_builder).args + fargs(model_builder).kwonlyargs


def list_layer_names(network, only_para_layers=True) -> List[str]:
    """Get a list of unique string representations of all layers in the network."""
    if only_para_layers:
        return [layer.name for layer in network.submodules if
                not isinstance(layer, (tf.keras.layers.Activation,
                                       tf.keras.layers.InputLayer,
                                       tf.keras.layers.Concatenate,
                                       tf.keras.layers.Reshape))]
    else:
        return [layer.name for layer in network.submodules if
                not isinstance(layer, tf.keras.layers.InputLayer)]


def get_layer_names(model: tf.keras.Model):
    """Get names of all (outer) layers in the model."""
    return [layer.name if not isinstance(layer, TimeDistributed) else layer.layer.name for layer in model.layers]


def extract_layers(network: tf.keras.Model, unfold_tds: bool = False) -> List[tf.keras.layers.Layer]:
    """Recursively extract layers from a potentially nested list of Sequentials of unknown depth."""
    if not hasattr(network, "layers"):
        return [network]

    layers = []
    for l in network.submodules:
        if isinstance(l, tf.keras.Model) or isinstance(l, tf.keras.Sequential):
            layers.append(extract_layers(l))
        elif isinstance(l, tf.keras.layers.TimeDistributed) and unfold_tds:
            if isinstance(l.layer, tf.keras.Model) or isinstance(l.layer, tf.keras.Sequential):
                layers.append(extract_layers(l.layer))
            else:
                layers.append(l.layer)
        else:
            layers.append(l)

    return flatten(layers)


def get_layers_by_names(network: tf.keras.Model, layer_names: List[str]):
    """Get a list of layers identified by their names from a network."""
    layers = network.submodules
    all_layer_names = [l.name for l in layers]

    assert all(ln in all_layer_names for ln in layer_names), \
        f"Cannot find layers {list(filter(lambda ln: ln not in all_layer_names, layer_names))} in network extraction."

    return [layers[all_layer_names.index(layer_name)] for layer_name in layer_names]


def get_parent_layer(child, network):
    if child in network.layers:
        return network

    for layer in network.submodules:
        if hasattr(layer, "layers") and child in layer.layers:
            return layer

    return None


def build_sub_model_to(network: tf.keras.Model, tos: Union[List[str], List[tf.keras.Model]], include_original=False):
    """Build a sub model of a given network that has (multiple) outputs at layer activations defined by a list of layer
    names."""
    layers = get_layers_by_names(network, tos) if isinstance(tos[0], str) else tos
    layer = layers[0]

    if not hasattr(layer, "layers") and layer in network.layers:
        return tf.keras.Model(inputs=network.inputs, outputs=layer.output)

    if hasattr(layer, "layers") and layer in network.layers:
        layer = layer.layers[-1]

    parent = get_parent_layer(layer, network)
    inbound_layers = parent.inbound_nodes[0].inbound_layers
    if not isinstance(inbound_layers, list):
        inbound_layers = [inbound_layers]

    outputs = [il.output if not hasattr(il, "layers") else build_sub_model_to(network, [il])(network.inputs) for il in
               inbound_layers]

    subnetwork_to_parent = tf.keras.Model(inputs=network.inputs, outputs=outputs)
    subnetwork_to_layer = tf.keras.Model(inputs=parent.inputs, outputs=layer.output)

    return tf.keras.Model(inputs=subnetwork_to_parent.inputs, outputs=subnetwork_to_layer(subnetwork_to_parent.output))


def build_sub_model_from(network: tf.keras.Model, from_layer_name: str):
    """EXPERIMENTAL: NOT GUARANTEED TO WORK WITH ALL ARCHITECTURES.

    Build a sub model of a given network that has inputs at layers defined by a given name.
    Outputs will remain the network outputs."""
    all_layer_names = [l.name for l in network.layers]

    first_layer_index = all_layer_names.index(from_layer_name)
    first_layer = network.layers[first_layer_index]

    new_input = tf.keras.layers.Input(first_layer.input_shape[1:])
    x = first_layer(new_input)
    for layer in network.layers[first_layer_index + 1:]:
        x = layer(x)

    return tf.keras.Model(inputs=new_input, outputs=x)


def get_component(model: tf.keras.Model, name: str):
    """Get outer layer/component by name."""
    for layer in model.layers:
        layer_name = layer.name
        if isinstance(layer, TimeDistributed):
            layer_name = layer.layer.name

        if layer_name == name:
            return layer


@tf.function
def reset_states_masked_tf(recurrent_layers: List[tf.keras.layers.RNN], mask: List):
    """Reset a stateful model's states only at the samples in the batch that are specified by the mask.

    The mask should be a list of length 'batch size' and contain one at every position where the state should be reset,
    and zeros otherwise (booleans possible too)."""

    for layer in recurrent_layers:
        current_states = layer.states
        new_states = []
        # new_states = tf.TensorArray(tf.float32, size=len(current_states))
        for i_state in range(len(current_states)):
            current_state = current_states[i_state]
            expanded_mask = tf.tile(
                tf.transpose(
                    tf.expand_dims(mask, axis=0)
                ),(1, current_state.shape[-1])
            )  # TODO why did we reverse?

            masked_reset_state = tf.where(expanded_mask, 0., current_state)
            new_states.append(masked_reset_state)
            # new_states.write(i_state, masked_reset_state)

        # we need to do this by hand here because Keras' reset_states() method does not work with tf.function
        flat_states = tf.nest.flatten(layer.states)
        for i, state in enumerate(flat_states):
            state.assign(new_states[i])

def reset_states_masked(recurrent_layers: List[tf.keras.Model], mask: List):
    """Reset a stateful model's states only at the samples in the batch that are specified by the mask.

    The mask should be a list of length 'batch size' and contain one at every position where the state should be reset,
    and zeros otherwise (booleans possible too)."""
    for layer in recurrent_layers:
        current_states = [state.numpy() for state in layer.states]
        initial_states = 0
        new_states = []
        for current_state in current_states:
            expanded_mask = numpy.tile(numpy.flip(numpy.rot90(numpy.expand_dims(mask, axis=0))), (1, current_state.shape[-1]))
            masked_reset_state = np.where(expanded_mask, initial_states, current_state)
            new_states.append(masked_reset_state)

        layer.reset_states(new_states)


def calc_max_memory_usage(model: tf.keras.Model):
    """Calculate memory requirement of a model per sample in bits."""
    layers = model.submodules
    n_shapes = int(numpy.sum(
        [numpy.prod(numpy.array([s if isinstance(s, int) else 1 for s in l.output_shape])) for l in layers]))
    n_parameters = model.count_params()

    # memory needed for saving activations during gradient calculation
    n_activations = 0
    for l in layers:
        if len(l.trainable_variables) == 0 or l.output_shape is None:
            continue

        activation_shapes = l.output_shape
        if not isinstance(activation_shapes[0], tuple):
            activation_shapes = [tuple(activation_shapes)]

    total_memory = (n_shapes + n_parameters + n_activations) * 32

    return total_memory * 1.1641532182693481 * 10 ** -10


CONVOLUTION_BASE_CLASS = tf.keras.layers.Conv2D.__bases__[0]


def is_conv(layer):
    """Check if layer is convolutional."""
    return isinstance(layer, CONVOLUTION_BASE_CLASS)


# Building helpers

def make_input_layers(env, bs, sequence_length=None):
    """Build input layers for a model acting on the given environment.

    If the model is not recurrent, provide None for the sequence_length (default)"""

    state_dimensionality, n_actions = env_extract_dims(env)

    inputs = []

    for modality in state_dimensionality.keys():
        shape = (bs,)
        if sequence_length is not None:
            shape += (sequence_length,)
        shape += tuple(state_dimensionality[modality])

        the_input = tf.keras.Input(batch_shape=shape, name=modality)
        inputs.append(the_input)

    return inputs


# Validators

def validate_model_builder(model_function: Callable, env: TaskWrapper, distribution: "BasePolicyDistribution") -> None:
    models = model_function(env, distribution)

    # validate model order
    if not models[1].output_shape[-1] == 1:
        raise ValueError("The model function returns a model tuple whose value network's output is not one-dimensional."
                         "Make sure your value function returns a valid one-dimensional scalar output and that the "
                         "model builder returns this network as the second element of the model-triple.")
    if not models[2].output_shape:
        pass


def validate_env_model_compatibility(env: TaskWrapper, model: tf.keras.Model) -> bool:
    """Validate whether the envs states can be processed by the model."""
    model_inputs = [ip.name for ip in model.inputs]
    env_outputs = list(env.reset()[0].dict().keys())

    if not sorted(model_inputs) == sorted(env_outputs):
        raise IncompatibleModelException(
            f"The model with inputs {model_inputs} cannot handle this environment's states with senses {env_outputs}."
        )

    return True



import abc

from dm_control import mjcf

from angorapy.tasks.world_building.entities import Robot


class PhysicalWorld(abc.ABC):
    """A physical world in which an environment is simulated."""

    @property
    @abc.abstractmethod
    def robot(self) -> Robot:
        ...

    @property
    @abc.abstractmethod
    def stage(self) -> Robot:
        ...

    @property
    def root(self) -> mjcf.RootElement:
        """Returns the root element of the physical world."""
        return self.stage.mjcf_model



#!/usr/bin/env python
"""Tools to visualize the weights/filters/neurons in a (convolutional) network."""
import math
import os
from typing import List, Iterable, Union

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import numpy
import tensorflow as tf
import tensorflow_addons as tfa
import tensorflow_datasets as tfds
from sklearn.manifold import TSNE
from tqdm import tqdm

from analysis.investigation import Investigator
from angorapy.pretrain import top_5_accuracy
from utilities.model_utils import extract_layers, build_sub_model_to, is_conv
from utilities.plotting import plot_image_tiling, transparent_cmap
from utilities.util import normalize


class Visualizer(Investigator):
    """Utilities for visualizing network layers/filters."""

    def __init__(self, network: tf.keras.Model, mode: str = "show"):
        """

        Args:
            network: the netowrk to be visualized
            mode: either one of [show, save, return], where show creates popups for results and save saves them into the
                figure directory, return does neither and only returns result
        """
        super().__init__(network, None)
        self.network = network
        self.mode = "show"
        self.set_mode(mode)

        self.figure_directory = "../docs/analysis/figures/"
        os.makedirs(self.figure_directory, exist_ok=True)

    def set_mode(self, mode: str):
        """ Set the mode of visualizations to either "show" or "save".

        Args:
            mode: either one of [show, save, return], where show creates popups for results and save saves them into the
                figure directory, return does neither and only returns result
        """
        assert mode in ["show", "save", "return"], "Illegal Analyzer Mode. Choose from show, save."
        self.mode = mode

    def numb_features_in_layer(self, layer_name: str) -> int:
        """Get the number of features/filters in the layer specified by its unique string representation."""
        return self.get_layer_by_name(layer_name).output_shape[-1]

    def list_layer_names(self, only_para_layers=False) -> List[str]:
        """Get a list of unique string representations of all layers in the network."""
        if only_para_layers:
            return [layer.name for layer in extract_layers(self.network) if
                    not isinstance(layer, tf.keras.layers.Activation)]
        else:
            return [layer.name for layer in extract_layers(self.network)]

    def layer_weights_plot(self, layer_name):
        """Plot the weights (ignoring the bias) of a convolutional layer as a tessellation of its filters."""
        assert layer_name in self.list_convolutional_layer_names()
        layer = extract_layers(self.network)[self.list_layer_names().index(layer_name)]

        weights = normalize(layer.get_weights()[0])
        filters = [weights[:, :, :, f] for f in range(layer.output_shape[-1])]
        plot_image_tiling(filters)
        plt.show() if self.mode == "show" else plt.savefig(f"{self.figure_directory}/weights_{layer_name}.pdf",
                                                           format="pdf")

        plt.close()

    def preferred_stimulus(self, layer_name: str, feature_ids: List[int] = None,
                           optimization_steps: int = 500) -> List:
        """Visualize a given layer by performing feature maximization.

        Explanation at: http://cs231n.stanford.edu/slides/2017/cs231n_2017_lecture12.pdf

        In feature maximization, a random input image is created (here, uniform noise) and then successively updated
        to maximize the response in a filter (feature) of the specified layer(serialization). The procedure is extended to start
        with a low-resolution input image that is upscaled after every 30 optimization steps by some factor. This
        encourages the optimization to go towards a local minimum with low-frequency patterns, which are generally
        easier to interpret.

        Args:
            layer_name (str): the unique string identifying the layer who'serialization response shall be maximized
            feature_ids (list): a list of filter/feature IDs that will be maximized and shown as a tiling
            optimization_steps: the number of optimization steps between upscalings.

        Returns:
            None
        """
        # build model that only goes to the layer of interest

        # obtain output layer and change its activation to linear
        layer = self.get_layer_by_name(layer_name)

        if "activation" in layer.get_config().keys() and layer.get_config()["activation"] == "softmax":
            layer_input_shape = (1,) + layer.input_shape[1:]
            layer_config = tf.keras.layers.serialize(layer)
            layer_config["config"]["activation"] = "linear"
            new_layer = tf.keras.layers.deserialize(layer_config)
            new_layer(tf.random.normal(layer_input_shape))
            new_layer.set_weights(layer.get_weights())

            # create a new model with the linear output
            prev_layer = self.get_layer_by_name(self.list_layer_names(only_para_layers=True)[-2])
            new_layer_output = new_layer(prev_layer.output)
            intermediate_model = tf.keras.Model(inputs=self.network.input, outputs=new_layer_output)
        else:
            intermediate_model = tf.keras.Model(inputs=self.network.input, outputs=layer.output) # build_sub_model_to(self.network, [layer])

        # intermediate_model = tf.keras.Model(inputs=self.network.input, outputs=layer.output)
        width_in, height_in = intermediate_model.input_shape[1], intermediate_model.input_shape[2]

        feature_ids = list(range(layer.output_shape[-1])) if feature_ids is None or feature_ids == [] else feature_ids
        feature_maximizations = []
        for feature_id in tqdm(feature_ids, desc="Feature Maximization", disable=True):
            image = tf.zeros((1, width_in, height_in, 3)) + 0.5 + tf.random.normal((1, width_in, height_in, 3), 0, 0.03)

            optimizer = tf.keras.optimizers.Adam(0.1)
            image_variable = tf.Variable(image, trainable=True)

            for i in tqdm(range(optimization_steps), desc=f"{layer_name} {feature_id}"):

                with tf.GradientTape() as tape:
                    activations = intermediate_model(image_variable)
                    if isinstance(activations, list):
                        activations = activations[0]
                    loss = - tf.reduce_mean(activations[:, :, :, feature_id]) if is_conv(layer) else activations[:,
                                                                                                     feature_id]
                    # add l2 norm regularization
                    loss += 0.0001 * tf.reduce_sum(tf.square(image))

                gradient = tape.gradient(loss, image_variable)
                optimizer.apply_gradients([(gradient, image_variable)])

                image_tensor = tf.convert_to_tensor(image_variable)
                if i < optimization_steps - 1 and i % 4 == 0:
                    image_tensor = tfa.image.mean_filter2d(image_tensor, 3, padding="REFLECT")

                image_variable = tf.Variable(image_tensor, trainable=True)

            image = tf.convert_to_tensor(image_variable)
            final_image = normalize(tf.squeeze(image).numpy())
            feature_maximizations.append(final_image)

        plt.clf()
        plot_image_tiling(feature_maximizations)

        if self.mode == "show":
            plt.show()
        elif self.mode == "save":
            plt.savefig(
                f"{self.figure_directory}/feature_maximization_{layer_name}_{'_'.join(map(str, feature_ids))}.pdf",
                format="pdf")

        return feature_maximizations

    def feature_map(self, layer_name: str, reference_img: numpy.ndarray, mode: str = "gray"):
        """Visualize the activation map of a given layer, either as a gray level image, as a heatmap on top of the
        original image, or as a bar plot.

        Args:
            layer_name (str): the unique string identifying the layer from which to draw the activations
            reference_img: the image serving as the input producing the activation
            mode: the mode of the visualization, either one of [gray, heat, plot]

        Returns:
            None
        """
        assert mode in ["gray", "grey", "heat", "plot"]

        layer = self.get_layer_by_name(layer_name)
        sub_model = tf.keras.Model(inputs=self.network.input, outputs=layer.output)

        reference_img = normalize(reference_img).astype(numpy.float32)
        reference_img = tf.expand_dims(reference_img, axis=0) if len(reference_img.shape) == 3 else reference_img
        reference_width, reference_height = reference_img.shape[1], reference_img.shape[2]

        feature_maps = sub_model(reference_img)
        n_filters = feature_maps.shape[-1]

        plt.clf()
        if mode == "heat":
            feature_maps = tf.image.resize(feature_maps, size=(reference_width, reference_height), method="bicubic")
            feature_maps = tf.squeeze(feature_maps)

            # prepare subplots
            tiles_per_row = math.ceil(math.sqrt(n_filters))
            fig, axes = plt.subplots(tiles_per_row, tiles_per_row)
            plt.subplots_adjust(wspace=0.05, hspace=0.05)
            hm_cmap = transparent_cmap(plt.cm.get_cmap("hot"))

            i = 0
            for row_of_axes in axes:
                for axis in row_of_axes:
                    if i < n_filters:
                        axis.imshow(tf.squeeze(tf.image.rgb_to_grayscale(reference_img)), cmap="gray")
                        axis.contourf(feature_maps[:, :, i], cmap=hm_cmap)
                    else:
                        axis.axis("off")
                    axis.set_xticks([])
                    axis.set_yticks([])
                    i += 1

        elif mode in ["gray", "grey"]:
            feature_maps = feature_maps.numpy()
            fig, axes = plot_image_tiling(
                [numpy.squeeze(feature_maps[:, :, :, i]) for i in range(layer.output_shape[-1])],
                cmap="gray")
        elif mode == "plot":
            feature_maps = tf.squeeze(feature_maps)
            mean_filter_responses = tf.reduce_mean(feature_maps, axis=[0, 1])
            fig, axes = plt.subplots()
            axes.bar(list(range(n_filters)), mean_filter_responses)

        else:
            raise NotImplementedError("WhAT iS thIS?")

        output_format = "png" if mode in ["heat", "gray", "grey"] else "pdf"

        if self.mode == "show":
            plt.show()
        elif self.mode == "save":
            plt.savefig(f"{self.figure_directory}/feature_maps_{layer_name}{f'_{mode}'}.{output_format}",
                        format=output_format, dpi=300)

        plt.clf()
        plt.cla()

        return feature_maps

    def saliency_map(self, reference: Union[numpy.ndarray, tf.Tensor], layer_name=None, neuron: int = None) -> None:
        """Create a saliency map indicating the importance of each input pixel for the output, based on gradients.
        Only works for fully connected layers!

        Explanation at: http://cs231n.stanford.edu/slides/2017/cs231n_2017_lecture12.pdf

        Args:
            reference: the image serving as a reference image
            layer_name: string identifier of the layer in which a neuron will be analyzed, if None, use last layer
            neuron: neuron index in the layer, if None use the one with the highest activity
        """

        # resize image to fit network input shape
        reference = tf.Variable(tf.image.resize(reference, size=self.network.input_shape[1:-1]) / 255)

        if layer_name is None:
            layer = self.get_layer_by_name(self.list_layer_names(only_para_layers=True)[-1])
            layer_index = -1
        else:
            layer = self.get_layer_by_name(layer_name)
            layer_index = self.list_layer_names(only_para_layers=True).index(layer_name)

        if is_conv(layer):
            raise ValueError("Cannot do saliency maps for convolutional layers!")

        if "activation" in layer.get_config().keys() and layer.get_config()["activation"] == "softmax":
            layer_input_shape = (1,) + layer.input_shape[1:]
            layer_config = tf.keras.layers.serialize(layer)
            layer_config["config"]["activation"] = "linear"
            new_layer = tf.keras.layers.deserialize(layer_config)
            new_layer(tf.random.normal(layer_input_shape))
            new_layer.set_weights(layer.get_weights())

            # create a new model with the linear output
            prev_layer = self.get_layer_by_name(self.list_layer_names(only_para_layers=True)[layer_index - 1])
            new_layer_output = new_layer(prev_layer.output)
            intermediate_model = tf.keras.Model(inputs=self.network.input, outputs=new_layer_output)
        else:
            intermediate_model = build_sub_model_to(self.network, [layer])

        # make prediction based on new layer
        with tf.GradientTape() as tape:
            linear_prediction = intermediate_model(tf.dtypes.cast(tf.expand_dims(reference, axis=0), dtype=tf.float32))
            if isinstance(linear_prediction, list):
                linear_prediction = linear_prediction[0]

            if neuron is None:
                max_feature = int(tf.argmax(linear_prediction, axis=1).numpy().item())
            else:
                max_feature = neuron
            loss = tf.sigmoid(linear_prediction[:, max_feature])

        output_gradient = tape.gradient(loss, reference)
        saliency_map = tf.reduce_max(tf.math.abs(output_gradient), axis=-1)

        if self.mode != "return":
            plt.imshow(saliency_map, cmap="jet")

        if self.mode == "show":
            plt.show()
        elif self.mode == "save":
            plt.savefig(f"{self.figure_directory}/saliency_map.pdf", format="pdf", dpi=300)

        return saliency_map

    def cluster_inputs(self, layer_name: str, input_images: List[numpy.ndarray], classes: Iterable[int]) -> None:
        """Cluster inputs based on the representation produced by a given layer, using the t-SNE algorithm.

        Args:
            layer_name (str): the unique string identifying the layer from which to take the representation.
            input_images (list): a list of images that will be clustered
            classes (list): the classes/categories of the images based on which they are colored in the cluster plot
        """
        layer = self.get_layer_by_name(layer_name)
        sub_model = tf.keras.Model(inputs=self.network.input, outputs=layer.output)

        # retrieve representations
        input_images = [tf.expand_dims(normalize(img).astype(numpy.float32), axis=0) for img in input_images]
        representations = [tf.reshape(sub_model(img), [-1]).numpy() for img in input_images]

        # cluster
        t_sne = TSNE()
        sneezed_datapoints = t_sne.fit_transform(representations)
        x, y = numpy.split(sneezed_datapoints, 2, axis=1)

        plt.clf()
        plt.scatter(x=numpy.squeeze(x), y=numpy.squeeze(y), c=classes, cmap="Paired")

        plt.show() if self.mode == "show" else plt.savefig(
            f"{self.figure_directory}/input_cluster_{layer_name}.pdf", format="pdf", dpi=300)

    @staticmethod
    def from_saved_model(model_path: str, mode: str = "show"):
        """Build the analyzer from a model path."""
        assert os.path.exists(model_path), "Model Path does not exist!"
        return Visualizer(tf.keras.models.load_model(model_path), mode=mode)


if __name__ == "__main__":
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    f_weights = False
    f_max = False
    f_map = False
    f_salience = True
    do_tsne = False

    tf.random.set_seed(69420)

    model = tf.keras.models.load_model("../storage/pretrained/visual_r.h5",
                                       custom_objects={"top_5_accuracy": top_5_accuracy})
    analyzer = Visualizer(model, mode="show")

    # WEIGHT VISUALIZATION
    if f_weights:
        analyzer.layer_weights_plot("conv2d")

    # FEATURE MAXIMIZATION
    if f_max:
        fids = [5]
        analyzer.preferred_stimulus("conv2d_4", feature_ids=fids)

    # FEATURE MAPS
    if f_map:
        layer = "conv2d_3"
        reference = mpimg.imread("img/hand.png")
        analyzer.feature_map(layer, reference, mode="gray")
        analyzer.feature_map(layer, reference, mode="heat")
        analyzer.feature_map(layer, reference, mode="plot")

    # SALIENCE
    if f_salience:
        reference = mpimg.imread("img/hand.png")
        for i in range(10):
            analyzer.saliency_map(reference, layer_name="dense_1", neuron=i)

    # T-SNE CLUSTERING
    if do_tsne:
        data, _ = tfds.load("cifar10", shuffle_files=True).values()
        data = data.map(lambda img: (tf.image.resize(img["image"], (224, 224)) / 255, img["label"]))
        data = list(iter(data.take(100)))
        images = [d[0] for d in data]
        classes = [d[1].numpy().item() for d in data]

        analyzer.cluster_inputs("predictions", images, classes)



import re

import numpy as np
from matplotlib import pyplot as plt

with open("results/predictability_10000.txt", "r") as f:
    results = [re.match("(.*) from (.*) has an R2 of (.*)", line) for line in f.readlines()]

results = [r.groups() for r in results if r is not None]
result_dict = {}

included_regions = ["M1_activation", "PMC_activation"]
included_targets = ["translation", "translation_to_10", "translation_to_50"]

for r_target, r_activity, r2 in results:
    if r_activity not in result_dict:
        result_dict[r_activity] = {}

    if r_activity not in included_regions:
        continue

    if r_target not in included_targets:
        continue

    result_dict[r_activity][r_target] = float(r2[:-1])

x = np.arange(len(included_targets))
width = 0.35

rects = []
fig, ax = plt.subplots()
ax.bar(x - width / 2,
       [result_dict["M1_activation"][t] for t in included_targets],
       width, label="M1")
ax.bar(x + width / 2,
       [result_dict["PMC_activation"][t] for t in included_targets],
       width, label="PMC")

ax.set_xticks(x, ["next orientation", "orientation in 80ms", "orientation in 400ms"])
ax.legend()
ax.set_ylabel("Predictability (R2)")

ax.set_ylim(0, 1)

# plt.xticks(rotation="vertical")
# plt.gcf().subplots_adjust(bottom=0.3)
plt.show()



import os
import random
from typing import List

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import tensorflow as tf
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from analysis.visualization import Visualizer
from utilities.plotting import transparent_cmap

os.chdir("../../../")

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

tf.random.set_seed(69420)

# load the input image
reference = mpimg.imread("analysis/img/hand.png")

# ANALYZERS
analyzer_c = Visualizer.from_saved_model("storage/pretrained/tanh_c.h5", mode="return")
analyzer_r = Visualizer.from_saved_model("storage/pretrained/tanh_r.h5", mode="return")
analyzer_p = Visualizer.from_saved_model("storage/pretrained/tanh_h.h5", mode="return")

axs: List[Axes]
fig: Figure = plt.figure(figsize=(9, 3))
grid = plt.GridSpec(2, 6)

ax: Axes
for row in [0, 1]:
    for col in [0, 1, 2, 3, 4, 5]:
        analyzer = analyzer_c
        if col > 1:
            analyzer = analyzer_r
        if col > 3:
            analyzer = analyzer_p

        salience = analyzer.saliency_map(reference, layer_name="dense_1", neuron=random.randint(0, 512))

        print([row, col])
        ax = fig.add_subplot(grid[row, col])
        ax.imshow(salience, cmap="jet")
        ax.set_xticks([])
        ax.set_yticks([])

plt.savefig("docs/figures/pretraining_methods_saliency.pdf", format="pdf", bbox_inches='tight')
plt.show()



from os import wait

import mujoco
from matplotlib import pyplot as plt

from angorapy import make_task
from mujoco import viewer
from PIL import Image

env = make_task("ManipulateBlockDiscreteAsynchronous-v0", render_mode="human")
env.world.robot.show_palm_site()
env.set_delta_t_simulation(0.002)
env.set_original_n_substeps_to_sspcs()
env.change_color_scheme("default")

# viewer.launch(mujoco.MjModel.from_xml_string(env.world.stage.mjcf_model.to_xml_string(), assets=env.world.stage.mjcf_model.get_assets()))
# viewer.launch(env.model, env.data)

while True:
    env.reset()
    done = False
    step = 0

    while not done:
        o, r, t, t2, i = env.step(env.action_space.sample())
        done = t or t2
        # if step % 10 == 0:
        #     plt.imshow(o.vision)
        #     plt.show()

        step += 1


import json
import re

import numpy as np
from matplotlib import pyplot as plt
import matplotlib

from angorapy.common.const import QUALITATIVE_COLOR_PALETTE as COLORS

font = {'family' : 'Times',
        'weight' : 'normal',
        'size'   : 12}

matplotlib.rc('font', **font)


with open("storage/predictability_repeated_august19d.json", "r") as f:
    results = json.load(f)



included_regions = ["m1_internal", "pmc_recurrent_layer", "goal"]
included_targets = ["translation", "translation_to_10", "translation_to_20", "translation_to_30", "translation_to_40", "translation_to_50"]
# included_regions = ["M1_activation", "pmc_recurrent_layer"]
# included_targets = ["reward", "goal"]
# included_regions = ["M1_activation", "pmc_recurrent_layer"]
# included_targets = ['rotation_matrix', 'rotation_matrix_last_10']


x = [8 * x for x in [1, 10, 20, 30, 40, 50]]

m1_means = [np.mean(results["m1_internal"][t]) for t in included_targets]
pmc_means = [np.mean(results["pmc_recurrent_layer"][t]) for t in included_targets]
goal_means = [np.mean(results["goal"][t]) for t in included_targets]

m1_stds = [np.std(results["m1_internal"][t]) for t in included_targets]
pmc_stds = [np.std(results["pmc_recurrent_layer"][t]) for t in included_targets]
goal_stds = [np.std(results["goal"][t]) for t in included_targets]

rects = []
fig, ax = plt.subplots()

ax.plot(x, m1_means, label="M1", marker=None, color=COLORS[0])
ax.plot(x, pmc_means, label="PMC", marker=None, color=COLORS[1])
ax.plot(x, goal_means, label="Goal", marker=None, linestyle="--", alpha=0.5, color=COLORS[2])

# add y error bars in same colors as mean bars
ax.errorbar(x, m1_means, yerr=m1_stds, fmt='none', ecolor=COLORS[0], capsize=2)
ax.errorbar(x, pmc_means, yerr=pmc_stds, fmt='none', ecolor=COLORS[1], capsize=2)

ax.set_xlabel("Orientation in X ms")
ax.legend(ncol=3)

ax.set_ylabel("Predictability (R2)")

ax.set_ylim(0, 1)

# plt.show()

# set figure size to 8x2 inches
fig.set_size_inches(8, 3)
plt.tight_layout()

# save the figure to a pdf file without whitespace around it
plt.savefig(f"predictability_future_translation.pdf", format="pdf", bbox_inches="tight")



#!/usr/bin/env python
"""Hybrid policy networks that utilize both visual and unstructured input data."""
import os

import gymnasium as gym
import keras_cortex
import tensorflow as tf
from tensorflow.keras.layers import TimeDistributed as TD
# from tensorflow_core.python.keras.utils import plot_model
from tensorflow.python.keras.utils.vis_utils import plot_model

from angorapy.common.const import VISION_WH
from angorapy.common.policies import BasePolicyDistribution, MultiCategoricalPolicyDistribution
from angorapy.models import register_model
from angorapy.utilities.core import env_extract_dims


def build_ssc_module(batch_and_sequence_shape, touch_input_shape, activation=tf.keras.layers.ReLU):
    """Build the model of the SSC."""

    ssc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + touch_input_shape, name="SSCInput")
    ssc = TD(tf.keras.layers.Dense(128, name="SSC_1"), name="TD_ssc_1")(ssc_input)
    ssc = activation(name="SSC_activation_1")(ssc)
    ssc = TD(tf.keras.layers.Dense(64, name="SSC_2"), name="TD_ssc_2")(ssc)
    ssc = activation(name="SSC_activation_2")(ssc)

    return tf.keras.Model(inputs=ssc_input, outputs=ssc, name="SomatosensoryCortex")


def build_ppc_module(batch_and_sequence_shape, vc_input_shape, ssc_input_shape, activation=tf.keras.layers.ReLU):
    """Build the PPC model."""
    vc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + vc_input_shape, name="VCInput")
    ssc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ssc_input_shape, name="SSCInput")

    spl_input = tf.keras.layers.concatenate([vc_input, ssc_input])

    spl = TD(tf.keras.layers.Dense(256, name="SPL"), name="TD_SPL")(spl_input)
    spl = activation(name="SPL_activation")(spl)

    ipl = TD(tf.keras.layers.Dense(256, name="IPL"), name="TD_IPL")(spl)
    ipl = activation(name="IPL_activation")(ipl)

    ips_input = tf.keras.layers.concatenate([ipl, ssc_input])
    ips = TD(tf.keras.layers.Dense(128, name="IPS"), name="TD_IPS")(ips_input)
    ips = activation(name="IPS_activation")(ips)


    return tf.keras.Model(inputs=[vc_input, ssc_input], outputs=[spl, ipl, ips], name="PosteriorParietalCortex")


def build_pfc_module(batch_and_sequence_shape, goal_input_shape, ssc_input_shape, it_input_shape, activation=tf.keras.layers.ReLU):
    """Build the PFC model. Returns output of MCC, LPFC."""

    goal_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + goal_input_shape, name="Goal Input")
    ssc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ssc_input_shape, name="SSC Input")
    it_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + it_input_shape, name="IT Input")

    mcc_input = tf.keras.layers.concatenate([goal_input, ssc_input])
    mcc = TD(tf.keras.layers.Dense(64, name="MCC"), name="TD_mcc")(mcc_input)
    mcc = activation(name="MCC_activation")(mcc)

    lpfc_input = tf.keras.layers.concatenate([mcc, goal_input, it_input])
    lpfc = TD(tf.keras.layers.Dense(128, name="LPFC"), name="TD_lpfc")(lpfc_input)
    lpfc = activation(name="LPFC_activation")(lpfc)


    return tf.keras.Model(inputs=[goal_input, ssc_input, it_input], outputs=[mcc, lpfc], name="PrefrontalCortex")


def build_mc_module(batch_and_sequence_shape, mcc_input_shape, lpfc_input_shape, ipl_input_shape,
                    ips_input_shape, ssc_input_shape, rnn_class=tf.keras.layers.LSTM, activation=tf.keras.layers.ReLU):
    mcc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + mcc_input_shape, name="GoalInput")
    lpfc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + lpfc_input_shape, name="LPFCInput")
    ipl_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ipl_input_shape, name="IPLInput")
    ips_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ips_input_shape, name="IPSInput")
    ssc_input = tf.keras.Input(batch_shape=batch_and_sequence_shape + ssc_input_shape, name="SSCInput")

    pmc_input = tf.keras.layers.concatenate([lpfc_input, ipl_input])

    pmc = TD(tf.keras.layers.Dense(512, name="PMC_dense"), name="TD_pmc_dense")(pmc_input)
    pmc = activation(name="PMC_activation")(pmc)

    pmc, *_ = rnn_class(512,
                        stateful=True,
                        return_sequences=True,
                        batch_size=batch_and_sequence_shape[0],
                        return_state=True,
                        name="pmc_recurrent_layer")(pmc)

    m1_input = tf.keras.layers.concatenate([pmc, mcc_input, lpfc_input, ssc_input, ips_input])

    m1 = TD(tf.keras.layers.Dense(256, name="m1"), )(m1_input)
    m1 = activation(name="M1_activation")(m1)

    return tf.keras.Model(inputs=[mcc_input, lpfc_input, ipl_input, ips_input, ssc_input],
                          outputs=[pmc, m1], name="MotorCortex")


@register_model("shadow")
def build_shadow_brain_base(env: gym.Env, distribution: BasePolicyDistribution, bs: int = 1, model_type: str = "rnn",
                            blind: bool = False, sequence_length=1, activation=tf.keras.layers.ReLU, **kwargs):
    """Build network for the shadow hand task, version 2."""
    state_dimensionality, n_actions = env_extract_dims(env)

    rnn_choice = {"rnn": tf.keras.layers.SimpleRNN,
                  "lstm": tf.keras.layers.LSTM,
                  "gru": tf.keras.layers.GRU}[model_type]

    # inputs
    visual_input = tf.keras.Input(batch_shape=(bs, sequence_length, *state_dimensionality["vision"],), name="vision")
    proprio_in = tf.keras.Input(batch_shape=(bs, sequence_length, *state_dimensionality["proprioception"],),
                                name="proprioception")
    touch_in = tf.keras.Input(batch_shape=(bs, sequence_length, *state_dimensionality["touch"],),
                              name="touch")
    goal_input = tf.keras.Input(batch_shape=(bs, sequence_length, *state_dimensionality["goal"],), name="goal")

    input_list = [visual_input, proprio_in, touch_in, goal_input]

    touch = tf.keras.layers.concatenate([proprio_in, touch_in])
    touch_masked = tf.keras.layers.Masking(
        batch_input_shape=touch.shape
    )(touch)

    if blind:
        vc = visual_input
    else:
        vc = TD(keras_cortex.cornet.CORNetZ(7), name="visual_component")(visual_input)

    vision_masked = tf.keras.layers.Masking(batch_input_shape=visual_input.shape)(vc)
    goal_masked = tf.keras.layers.Masking(batch_input_shape=goal_input.shape)(goal_input)

    ssc = build_ssc_module((bs, sequence_length,), (touch_masked.shape[-1],), activation=activation)
    ssc_out = ssc(touch_masked)

    ppc = build_ppc_module((bs, sequence_length,), (vision_masked.shape[-1],), (ssc.output_shape[-1],), activation=activation)
    spl_out, ipl_out, ips_out = ppc([vision_masked, ssc_out])

    pfc = build_pfc_module((bs, sequence_length,), (goal_masked.shape[-1],), (ssc.output_shape[-1],),
                           (vision_masked.shape[-1],), activation=activation)
    mcc_out, lpfc_out = pfc([goal_masked, ssc_out, vision_masked])

    mc = build_mc_module((bs, sequence_length,), (mcc_out.shape[-1],), (lpfc_out.shape[-1],), (ipl_out.shape[-1],),
                         (ips_out.shape[-1],), (ssc_out.shape[-1],), rnn_class=rnn_choice, activation=activation)
    pmc_out, m1_out = mc([mcc_out, lpfc_out, ipl_out, ips_out, ssc_out])

    # policy head
    policy_out = distribution.build_action_head(n_actions, m1_out.shape[1:], bs)(m1_out)

    # value head
    value_inputs = [pmc_out, mcc_out, ips_out, lpfc_out, ssc_out]
    if "asymmetric" in state_dimensionality.keys():
        asymmetric = tf.keras.Input(batch_shape=(bs, sequence_length,) + state_dimensionality["asymmetric"],
                                      name="asymmetric")
        input_list.append(asymmetric)
        value_inputs.append(asymmetric)

    value_in = tf.keras.layers.concatenate(value_inputs)
    value_out = tf.keras.layers.Dense(512)(value_in)
    value_out = activation()(value_out)
    value_out = tf.keras.layers.Dense(512)(value_out)
    value_out = activation()(value_out)
    value_out = tf.keras.layers.Dense(1, name="value")(value_out)

    # define models
    policy = tf.keras.Model(inputs=input_list, outputs=[policy_out], name="shadow_brain_policy")
    value = tf.keras.Model(inputs=input_list, outputs=[value_out], name="shadow_brain_value")
    joint = tf.keras.Model(inputs=input_list, outputs=[policy_out, value_out],
                           name="shadow_brain" + ("_visual" if not blind else ""))

    return policy, value, joint


if __name__ == "__main__":
    from angorapy import make_task

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
    # tf.config.experimental.set_memory_growth(tf.config.list_physical_devices("GPU")[0], True)

    batch_size = 16
    sequence_length = 16

    optimizer: tf.keras.optimizers.Optimizer = tf.keras.optimizers.SGD()

    # without vision
    env = make_task("HumanoidManipulateBlockDiscreteAsynchronous-v0")
    _, _, joint = build_shadow_brain_base(env, MultiCategoricalPolicyDistribution(env), bs=batch_size, blind=True,
                                          sequence_length=sequence_length, model_type="gru")
    plot_model(joint, to_file=f"{joint.name}.png", expand_nested=True, show_shapes=True)

    out = joint({
        "vision": tf.random.normal((batch_size, sequence_length, 7)),
        "proprioception": tf.random.normal((batch_size, sequence_length, 48)),
        "touch": tf.random.normal((batch_size, sequence_length, 92)),
        "asymmetric": tf.random.normal((batch_size, sequence_length, 25)),
        "goal": tf.random.normal((batch_size, sequence_length, 4)),
    })

    joint.summary()

    # with vision
    env = make_task("HumanoidVisualManipulateBlockDiscreteAsynchronous-v0")
    _, _, joint = build_shadow_brain_base(env, MultiCategoricalPolicyDistribution(env), bs=batch_size, blind=False,
                                          sequence_length=sequence_length, model_type="gru")
    plot_model(joint, to_file=f"{joint.name}.png", expand_nested=True, show_shapes=True)

    out = joint({
        "vision": tf.random.normal((batch_size, sequence_length, VISION_WH, VISION_WH, 3)),
        "proprioception": tf.random.normal((batch_size, sequence_length, 48)),
        "touch": tf.random.normal((batch_size, sequence_length, 92)),
        "asymmetric": tf.random.normal((batch_size, sequence_length, 25)),
        "goal": tf.random.normal((batch_size, sequence_length, 4)),
    })

    joint.summary()



import os

import matplotlib.pyplot as plt
import seaborn as sns

from agent.ppo_agent import PPOAgent

AGENT_ID = 1582658038

os.chdir("../../../")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
agent = PPOAgent.from_agent_state(AGENT_ID, "b", path_modifier="")

# stats = StatBundle(10, 10, [3,6,3,8,6,7,4,3,6,7,4], [], None)
stats = agent.evaluate(100)

sns.set(style="whitegrid")
violin_ax = sns.violinplot(y=stats.episode_rewards)
swarm_ax = sns.swarmplot(y=stats.episode_rewards, color=".25")

plt.ylabel("Episode Reward")

plt.subplots_adjust(wspace=.5, bottom=0.2)
plt.savefig("docs/figures/reaching_evaluation.pdf", format="pdf", bbox_inches='tight')
plt.show()



#!/usr/bin/env python
"""Custom layer components to be used in model implementations."""
import tensorflow as tf


class StdevLayer(tf.keras.layers.Layer):
    """Layer carrying parameters to be directly optimized, serving as standard deviations in continuous gaussian
    policies."""

    def __init__(self, n_actions, **kwargs):
        super(StdevLayer, self).__init__(**kwargs)
        self.n_actions = n_actions

    def build(self, input_shape):
        """Build the layer by adding the parameters representing the standard deviations."""

        # initializing the log stdevs as zeros essentially means we initialize the true standard deviations to 1
        self.log_stdevs = self.add_weight("log_stdevs", shape=[self.n_actions, 1], initializer='zeros', trainable=True)

    def call(self, input, **kwargs):
        return tf.matmul(tf.ones_like(input), self.log_stdevs)


class BetaDistributionSpreadLayer(tf.keras.layers.Layer):
    """Layer carrying parameters to be directly optimized, serving as v = a + b values in reparameterized beta policies.
    """

    def __init__(self, n_actions, **kwargs):
        super().__init__(**kwargs)
        self.n_actions = n_actions

    def build(self, input_shape):
        """Build the layer by adding the parameters representing the standard deviations."""

        # initializing the log stdevs as zeros essentially means we initialize to 2
        self.spreads = self.add_weight("sum_ab", shape=[self.n_actions, 1], initializer='zeros', trainable=True)

    def call(self, input, **kwargs):
        return tf.matmul(tf.ones_like(input), self.spreads)



#!/usr/bin/env python
"""Collection of generic fully connected and recurrent policy networks."""

import os
from contextlib import suppress
from typing import Tuple

import gymnasium as gym
import tensorflow as tf
from tensorflow.keras.layers import TimeDistributed
from tensorflow.python.keras.utils.vis_utils import plot_model

from angorapy.common.policies import BasePolicyDistribution, CategoricalPolicyDistribution, BetaPolicyDistribution, \
    MultiCategoricalPolicyDistribution, RBetaPolicyDistribution, GaussianPolicyDistribution
from angorapy.models import register_model
from angorapy.tasks.wrappers import TaskWrapper
from angorapy.models.components import _build_encoding_sub_model
from angorapy.utilities.model_utils import make_input_layers
from angorapy.utilities.core import env_extract_dims


def build_ffn_models(env: TaskWrapper,
                     distribution: BasePolicyDistribution,
                     shared: bool = False,
                     layer_sizes: Tuple = (64, 64)):
    """Build a simple fully connected feed-forward model."""

    # preparation
    state_dimensionality, n_actions = env_extract_dims(env)
    input_list = make_input_layers(env, None, None)

    if len(input_list) > 1:
        policy_inputs = list(filter(lambda i: i.name != "asymmetric", input_list))
        inputs = tf.keras.layers.Concatenate(name="flat_inputs")(policy_inputs)
    else:
        inputs = input_list[0]

    # policy network
    latent = _build_encoding_sub_model(inputs.shape[1:], None, layer_sizes=layer_sizes, name="policy_encoder")(inputs)
    out_policy = distribution.build_action_head(n_actions, (layer_sizes[-1],), None)(latent)

    policy = tf.keras.Model(inputs=input_list, outputs=out_policy, name="policy")

    # value network
    if "asymmetric" in state_dimensionality.keys():
        inputs = tf.keras.layers.Concatenate(name="added_asymmetric_inputs")(
            [inputs, list(filter(lambda i: i.name == "asymmetric", input_list))[0]]
        )

    if not shared:
        value_latent = _build_encoding_sub_model(inputs.shape[1:], None, layer_sizes=layer_sizes, name="value_encoder")(inputs)
        value_out = tf.keras.layers.Dense(1, kernel_initializer=tf.keras.initializers.Orthogonal(1.0),
                                          bias_initializer=tf.keras.initializers.Constant(0.0))(value_latent)
    else:
        value_out = tf.keras.layers.Dense(1, input_dim=layer_sizes[-1],
                                          kernel_initializer=tf.keras.initializers.Orthogonal(1.0),
                                          bias_initializer=tf.keras.initializers.Constant(0.0))(latent)

    value = tf.keras.Model(inputs=input_list, outputs=value_out, name="value")

    return policy, value, tf.keras.Model(inputs=input_list, outputs=[out_policy, value_out], name="policy_value")


def build_rnn_models(env: TaskWrapper,
                     distribution: BasePolicyDistribution,
                     shared: bool = False,
                     bs: int = 1,
                     model_type: str = "gru",
                     layer_sizes: Tuple = (64, 64),
                     sequence_length=1):
    """Build simple policy and value models having a recurrent layer before their heads.

    Args:
        sequence_length:
    """

    state_dimensionality, n_actions = env_extract_dims(env)
    input_list = make_input_layers(env, bs=bs, sequence_length=sequence_length)

    if len(input_list) > 1:
        policy_inputs = list(filter(lambda i: i.name != "asymmetric", input_list))
        inputs = tf.keras.layers.Concatenate(name="flat_inputs")(policy_inputs)
    else:
        inputs = input_list[0]

    if "asymmetric" in state_dimensionality.keys():
        value_inputs = tf.keras.layers.Concatenate(name="added_asymmetric_inputs")(
            [inputs, list(filter(lambda i: i.name == "asymmetric", input_list))[0]]
        )
    else:
        value_inputs = inputs

    masked = tf.keras.layers.Masking(batch_input_shape=(bs, sequence_length,) + (inputs.shape[-1], ))(inputs)

    # policy network; stateful, so batch size needs to be known
    encoder_sub_model = _build_encoding_sub_model(
        inputs.shape[-1],
        # since we are distributing the encoder over all timesteps, we need to blow up the bs here
        bs * sequence_length,
        layer_sizes=layer_sizes[:-1],
        name="policy_encoder")

    x = TimeDistributed(encoder_sub_model, name="TD_policy")(masked)
    x.set_shape([bs] + x.shape[1:])

    rnn_choice = {"rnn": tf.keras.layers.SimpleRNN,
                  "lstm": tf.keras.layers.LSTM,
                  "gru": tf.keras.layers.GRU}[model_type]

    x, *_ = rnn_choice(layer_sizes[-1],
                       stateful=True,
                       return_sequences=True,
                       return_state=True,
                       batch_size=bs,
                       name="policy_recurrent_layer")(x)

    out_policy = distribution.build_action_head(n_actions, x.shape[1:], bs)(x)
    policy = tf.keras.Model(inputs=input_list, outputs=out_policy, name="simple_rnn_policy")

    # value network
    if not shared:
        value_inputs_masked = tf.keras.layers.Masking(batch_input_shape=(bs, sequence_length,) + (value_inputs.shape[-1],))(value_inputs)
        x = TimeDistributed(
            _build_encoding_sub_model(value_inputs_masked.shape[-1],
                                      bs * sequence_length,
                                      layer_sizes=layer_sizes[:-1],
                                      name="value_encoder"),
            name="TD_value")(value_inputs_masked)
        x.set_shape([bs] + x.shape[1:])

        x, *_ = rnn_choice(layer_sizes[-1], stateful=True, return_sequences=True, return_state=True, batch_size=bs,
                           name="value_recurrent_layer")(x)
        out_value = tf.keras.layers.Dense(1, kernel_initializer=tf.keras.initializers.Orthogonal(1.0),
                                          bias_initializer=tf.keras.initializers.Constant(0.0), name="value_out")(x)
    else:
        if "asymmetric" in state_dimensionality.keys():
            x = tf.keras.layers.Concatenate()([x, asymmetric_inputs])

        out_value = tf.keras.layers.Dense(1, input_dim=x.shape[1:],
                                          kernel_initializer=tf.keras.initializers.Orthogonal(1.0),
                                          bias_initializer=tf.keras.initializers.Constant(0.0), name="value_out")(x)

    value = tf.keras.Model(inputs=input_list, outputs=out_value, name="simple_rnn_value")

    return policy, value, tf.keras.Model(inputs=input_list, outputs=[out_policy, out_value], name="simple_rnn")


@register_model("simple")
def build_simple_models(env: TaskWrapper,
                        distribution: BasePolicyDistribution,
                        shared: bool = False,
                        bs: int = 1,
                        sequence_length: int = 1,
                        model_type: str = "gru",
                        **kwargs):
    """Build simple networks (policy, value, joint) for given parameter settings."""

    # this function is just a wrapper routing the requests for ffn and rnns
    if model_type == "ffn":
        return build_ffn_models(env, distribution, shared, layer_sizes=(64, 64))
    else:
        return build_rnn_models(env, distribution, shared, bs=bs, sequence_length=sequence_length, model_type=model_type,
                                layer_sizes=(64, 64))


@register_model("deeper")
def build_deeper_models(env: TaskWrapper,
                        distribution: BasePolicyDistribution,
                        shared: bool = False,
                        bs: int = 1,
                        sequence_length: int = 1,
                        model_type: str = "gru",
                        **kwargs):
    """Build deeper simple networks (policy, value, joint) for given parameter settings."""

    # this function is just a wrapper routing the requests for ffn and rnns
    if model_type == "ffn":
        return build_ffn_models(env, distribution, shared, layer_sizes=(64, 64, 64, 32))
    else:
        return build_rnn_models(env, distribution, shared, bs=bs, sequence_length=sequence_length,
                                model_type=model_type, layer_sizes=(64, 64, 64, 32, 32))


@register_model("wider")
def build_wider_models(env: TaskWrapper,
                       distribution: BasePolicyDistribution,
                       shared: bool = False,
                       bs: int = 1,
                       sequence_length: int = 1,
                       model_type: str = "gru",
                       **kwargs):
    """Build deeper simple networks (policy, value, joint) for given parameter settings."""

    # this function is just a wrapper routing the requests for ffn and rnns
    if model_type == "ffn":
        return build_ffn_models(env, distribution, shared, layer_sizes=(1024, 512))
    else:
        return build_rnn_models(env, distribution, shared, bs=bs, sequence_length=sequence_length,
                                model_type=model_type, layer_sizes=(1024, 512))


if __name__ == '__main__':
    from angorapy import make_task

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    cont_env = make_task("HumanoidManipulateBlockDiscreteAsynchronous-v0")
    discrete_env = gym.make("LunarLander-v2")
    multi_discrete_env = gym.make("ManipulateBlockDiscreteRelative-v0")

    # model = build_simple_models(cont_env, RBetaPolicyDistribution(cont_env), False, 1, 1, "gru")
    # print(f"Simple model has {model[0].count_params()} parameters.")
    # plot_model(model[2], show_shapes=True, to_file="simple.png", expand_nested=True)
    #
    # model = build_simple_models(discrete_env, CategoricalPolicyDistribution(discrete_env), False, 1, 1, "gru")
    # print(f"Simple model has {model[0].count_params()} parameters.")
    # plot_model(model[2], show_shapes=True, to_file="model_graph_simple_discrete.png", expand_nested=True)
    #
    # model = build_simple_models(multi_discrete_env, MultiCategoricalPolicyDistribution(multi_discrete_env), False, 1, 1, "gru")
    # print(f"Simple model has {model[0].count_params()} parameters.")
    # plot_model(model[2], show_shapes=True, to_file="model_graph_simple_multidiscrete.png", expand_nested=True)
    #
    # model = build_deeper_models(cont_env, BetaPolicyDistribution(cont_env), False, 1, 1, "gru")
    # print(f"Deeper model has {model[0].count_params()} parameters.")
    # plot_model(model[2], show_shapes=True, to_file="model_graph_deeper.png", expand_nested=True)

    model = build_wider_models(cont_env, MultiCategoricalPolicyDistribution(cont_env), False, 1, 1, "gru")
    print(f"Wider model without weight sharing has {model[2].count_params()} parameters.")
    plot_model(model[2], show_shapes=True, to_file="model_graph_wider_unshared.png", expand_nested=True)

    model = build_wider_models(cont_env, MultiCategoricalPolicyDistribution(cont_env), False, 1, 1, "gru")
    print(f"Wider model without weight sharing has {model[2].count_params()} parameters.")
    plot_model(model[2], show_shapes=True, to_file="model_graph_wider_unshared_asymmetric.png", expand_nested=True)

    model = build_wider_models(cont_env, MultiCategoricalPolicyDistribution(cont_env), True, 1, 1, "gru")
    print(f"Wider model with weight sharing has {model[2].count_params()} parameters.")
    plot_model(model[2], show_shapes=True, to_file="model_graph_wider_shared.png", expand_nested=True)



from typing import List

import numpy as np
import tqdm

from angorapy.tasks.envs.dexterity.consts import FINGERTIP_SITE_NAMES
from angorapy.tasks.envs.dexterity.utils import get_fingertip_distance


# HELPER FUNCTIONS

def calculate_force_penalty(data) -> float:
    """Calculate a penalty for applying force. Sum of squares of forces, ignoring the wrist."""
    return (data.actuator_force[2:] ** 2).sum()


def calculate_tendon_stress_penalty(data) -> float:
    """Calculate a penalty for applying force. Sum of squares of forces, ignoring the wrist. Scaled up by a constant
    to match force penalty in average scale."""
    return ((data.ten_length ** 2) * 116).sum()


def calculate_auxiliary_finger_penalty(environment, exclude: List[int] = None) -> float:
    """Calculate a penalty for auxiliary (non-target) fingers being in a zone around the thumb."""
    if exclude is None:
        exclude = []

    penalty = 0
    for i, fname in enumerate(FINGERTIP_SITE_NAMES):
        if fname == environment.thumb_name or i == np.where(environment.goal == 1)[0].item() or i in exclude:
            continue

        fingertip_distance = get_fingertip_distance(environment.get_thumb_position(),
                                                    environment.get_finger_position(fname))

        # cap reward at auxiliary zone, scale to focus on the target finger'serialization movement
        penalty += ((min(fingertip_distance, environment.reward_config["AUXILIARY_ZONE_RADIUS"])
                     - environment.reward_config[
                         "AUXILIARY_ZONE_RADIUS"])  # base reward is 0, being in the zone is punished
                    * environment.reward_config["AUXILIARY_PENALTY_MULTIPLIER"])

    return - penalty


# REWARD FUNCTIONS

def reach(env: "BaseShadowHandEnv", info: dict):
    """Simple reward function for reaching tasks, combining distance, force and success."""
    return (- get_fingertip_distance(env.get_fingertip_positions(), env.goal)
            + info["is_success"] * env.reward_config["SUCCESS_BONUS"]
            - calculate_force_penalty(env.data) * env.reward_config["FORCE_MULTIPLIER"])


def free_reach(env: "AnthropomorphicEnv", info: dict):
    """Reward the relative join of the thumb'serialization and a target'serialization fingertip while punishing close
    other fingertips."""

    thumb_position = env.get_thumb_position()

    reward = (- get_fingertip_distance(thumb_position, env.get_target_finger_position())
              + info["is_success"] * env.reward_config["SUCCESS_BONUS"]
              - calculate_force_penalty(env.data) * env.reward_config["FORCE_MULTIPLIER"])

    reward -= calculate_auxiliary_finger_penalty(env)

    return reward


def free_reach_positive_reinforcement(env: "AnthropomorphicEnv", info: dict):
    """Reward progress towards the goal position while punishing other fingers for interfering."""
    # positive reinforcement
    progress_reward = (
            get_fingertip_distance(env.get_thumbs_previous_position(), env.get_target_fingers_previous_position())
            - get_fingertip_distance(env.get_thumb_position(), env.get_target_finger_position()))
    success_reward = info["is_success"] * env.reward_config["SUCCESS_BONUS"]
    reinforcement_reward = progress_reward + success_reward

    # positive punishment
    penalty = (calculate_force_penalty(env.data) * env.reward_config["FORCE_MULTIPLIER"]
               + calculate_auxiliary_finger_penalty(env))

    # total reward
    return reinforcement_reward - penalty


# SEQUENTIAL REACHING

def sequential_reach(env: "BaseShadowHandEnv", info: dict):  # TODO adjust for sequence?
    """Reward following a sequence of finger configurations."""
    reward = (- get_fingertip_distance(env.get_fingertip_positions(), env.goal)
              + info["is_success"] * env.reward_config["SUCCESS_BONUS"]
              - calculate_force_penalty(env.data) * env.reward_config["FORCE_MULTIPLIER"])

    return reward


def sequential_free_reach(env: "BaseShadowHandEnv", info: dict):
    """Reward following a sequence of reach movements."""

    # reinforcement
    progress_reward = (
            get_fingertip_distance(env.get_thumbs_previous_position(), env.get_target_fingers_previous_position())
            - get_fingertip_distance(env.get_thumb_position(), env.get_target_finger_position()))
    success_reward = info["is_success"] * env.reward_config["SUCCESS_BONUS"]
    reinforcement_reward = progress_reward + success_reward

    # punishment
    exclusions = []
    if len(env.goal_sequence) > 1:
        exclusions = [env.current_target_finger]

    punishment_reward = (- calculate_force_penalty(env.data) * env.reward_config["FORCE_MULTIPLIER"]
                         - calculate_auxiliary_finger_penalty(env, exclude=exclusions))

    return reinforcement_reward + punishment_reward


# MANIPULATE

def manipulate(env: "BaseManipulate", info: dict):
    """Reward function fo object manipulation tasks."""
    success = env._is_success(env.get_object_pose(), env.goal).astype(np.float32)
    d_progress = env._goal_progress()

    reward = (+ d_progress  # convergence to goal reward
              + success * env.reward_config["SUCCESS_BONUS"]  # reward for finishing
              - env._is_dropped() * env.reward_config["DROPPING_PENALTY"])  # dropping penalty

    reward -= calculate_force_penalty(env.data) * env.reward_config["FORCE_MULTIPLIER"]
    reward -= calculate_tendon_stress_penalty(env.data) * env.reward_config["TENDON_STRESS_MULTIPLIER"]

    return reward


if __name__ == '__main__':
    from environments import *

    environment = gym.make("HumanoidManipulateBlockDiscreteAsynchronous-v0")
    environment.reset()

    force_penalties = []
    tendon_penalties = []
    for i in tqdm.tqdm(range(100000)):
        o, r, d, _ = environment.step(environment.action_space.sample())

        tendon_penalties.append(calculate_tendon_stress_penalty(environment.data))
        force_penalties.append(calculate_force_penalty(environment.data))

        if d:
            environment.reset()

    print(f"Mean Force Penalty: {np.mean(force_penalties)}")
    print(f"Mean Tendon Penalty: {np.mean(tendon_penalties)}")
    print(f"Suggested Scaling: {np.round(np.mean(force_penalties) / np.mean(tendon_penalties))}")



from gymnasium.spaces import Box, Discrete, MultiDiscrete

from angorapy.common.policies import get_distribution_by_short_name


def autoselect_distribution(env):
    """For a given environment, select a compatible default distribution."""
    if isinstance(env.action_space, Box):
        distribution = "gaussian"
    elif isinstance(env.action_space, Discrete):
        distribution = "categorical"
    elif isinstance(env.action_space, MultiDiscrete):
        distribution = "multi-categorical"
    else:
        raise NotImplementedError("No appropriate distribution found for environment.")

    return get_distribution_by_short_name(distribution)(env)



#!/usr/bin/env python
import copy
import os
from time import sleep
from typing import List, Union

import gymnasium as gym
import matplotlib.pyplot as plt
import tensorflow as tf

from angorapy.common.policies import BasePolicyDistribution
from angorapy.tasks.wrappers import TaskWrapper
from angorapy.agent.ppo_agent import PPOAgent
from angorapy.utilities.hooks import register_hook, clear_hooks
from angorapy.utilities.model_utils import is_recurrent_model, list_layer_names, get_layers_by_names, build_sub_model_to, \
    extract_layers, CONVOLUTION_BASE_CLASS, is_conv
from angorapy.utilities.core import add_state_dims, flatten, insert_unknown_shape_dimensions


class Investigator:
    """Interface for investigating a policy network.

    The class serves as a wrapper to use a collection of methods that can extract information about the network."""

    def __init__(self, network: tf.keras.Model, distribution: BasePolicyDistribution):
        """Build an investigator for a network using a distribution.

        Args:
            network (tf.keras.Model):               the policy network to investigate
            distribution (BasePolicyDistribution):     a distribution that the network predicts
        """
        self.network: tf.keras.Model = network
        self.distribution: BasePolicyDistribution = distribution
        self.is_recurrent = is_recurrent_model(self.network)

    @classmethod
    def from_agent(cls, agent: PPOAgent):
        """Instantiate an investigator from an agent object."""
        agent.policy, agent.value, agent.joint = agent.build_models(agent.joint.get_weights(),
                                                                    batch_size=1, sequence_length=1)
        return cls(agent.policy, agent.distribution)

    def list_layer_names(self, only_para_layers=True) -> List[str]:
        """Get a list of unique string representations of all layers in the network."""
        return list_layer_names(self.network, only_para_layers=only_para_layers)

    def list_convolutional_layer_names(self) -> List[str]:
        """Get a list of unique string representations of convolutional layers in the network."""
        return [layer.name for layer in extract_layers(self.network) if is_conv(layer)]

    def list_non_convolutional_layer_names(self) -> List[str]:
        """Get a list of unique string representations of non-convolutional layers in the network."""
        return [layer.name for layer in extract_layers(self.network) if
                not isinstance(layer, CONVOLUTION_BASE_CLASS) and not isinstance(layer, tf.keras.layers.Activation)]

    def get_layers_by_names(self, layer_names: Union[List[str], str]):
        """Retrieve the layer object identified from the model by its unique string representation."""
        if not isinstance(layer_names, list):
            layer_names = [layer_names]
        return get_layers_by_names(self.network, layer_names=layer_names)

    def get_layer_by_name(self, layer_name):
        """Retrieve the layer object identified from the model by its unique string representation."""
        return self.get_layers_by_names(layer_name)[0]

    def get_layer_weights(self, layer_name):
        """Get the weights of a layer identified by its unique name."""
        return self.get_layers_by_names(layer_name)[0].get_weights()

    def get_weight_dict(self):
        """Get a dictionary mapping layer names to the weights of the layers."""
        out = {}
        for layer_name in self.list_layer_names(only_para_layers=True):
            out[layer_name] = self.get_layer_weights(layer_name)

        return out

    def plot_model(self):
        """Plot the network graph into a file."""
        tf.keras.utils.plot_model(self.network, show_shapes=True)

    def dissect_recurrent_layer_weights(self, layer_name):
        """Return a recurrent cells weights and biases in a named dictionary."""
        layer = self.get_layers_by_names(layer_name)[0]

        if not is_recurrent_model(layer):
            raise ValueError("Cannot dissect non-recurrent layer...")

        units = layer.units

        # stolen from https://stackoverflow.com/a/51484524/5407682
        W = layer.get_weights()[0]
        U = layer.get_weights()[1]
        b = layer.get_weights()[2]

        if isinstance(layer, tf.keras.layers.SimpleRNN):
            return dict(
                W=W,
                U=U,
                b=b
            )
        elif isinstance(layer, tf.keras.layers.GRU):
            return dict(
                W_z=W[:, :units],
                W_r=W[:, units: units * 2],
                W_c=W[:, units * 2:],

                U_z=U[:, :units],
                U_r=U[:, units: units * 2],
                U_c=U[:, units * 2:],

                b_z=b[:units],
                b_r=b[units: units * 2],
                b_c=b[units * 2:],
            )
        elif isinstance(layer, tf.keras.layers.LSTM):
            return dict(
                W_i=W[:, :units],
                W_f=W[:, units: units * 2],
                W_c=W[:, units * 2: units * 3],
                W_o=W[:, units * 3:],

                U_i=U[:, :units],
                U_f=U[:, units: units * 2],
                U_c=U[:, units * 2: units * 3],
                U_o=U[:, units * 3:],

                b_i=b[:units],
                b_f=b[units: units * 2],
                b_c=b[units * 2: units * 3],
                b_o=b[units * 3:],
            )
        else:
            raise ValueError("Recurrent layer type not understood. Is it custom?")

    def get_layer_activations(self, layer_names: List[str], input_tensor=None, recurrent_strategy="first"):
        """Get activations of a layer. If no input tensor is given, a random tensor is used."""
        activations = {}

        def activation_hook(module, input, output):
            activations[module.name] = output

        register_hook(get_layers_by_names(self.network, layer_names), after_call=activation_hook)
        if input_tensor is None:
            input_tensor = tf.random.normal(insert_unknown_shape_dimensions(self.network.input_shape))

        output = self.network(input_tensor, training=False)
        activations["output"] = output

        for key in activations.keys():
            if isinstance(activations[key], list):
                if recurrent_strategy == "first":
                    activations[key] = activations[key][0]
                elif recurrent_strategy == "all":
                    activations[key] = tf.concat(activations[key][1:], axis=-1)

        # release the hook to prevent infinite nesting
        clear_hooks(self.network)

        return activations

    def get_activations_over_episode(self, layer_names: Union[List[str], str], env: gym.Env, render: bool = False):
        """Run an episode using the network and get (serialization, activation, r) tuples for each timestep."""
        layer_names = layer_names if isinstance(layer_names, list) else [layer_names]

        states, activations, reward_trajectory, action_trajectory = [], [], [], []

        # make new model with multiple outputs
        polymodel = build_sub_model_to(self.network, layer_names, include_original=True)
        is_recurrent = is_recurrent_model(self.network)

        done = False
        state = env.reset()
        env.goal = 1
        env.render() if render else ""
        while not done:
            dual_out = flatten(polymodel.predict(add_state_dims(parse_state(state), dims=2 if is_recurrent else 1)))
            activation, probabilities = dual_out[:-len(self.network.output)], dual_out[-len(self.network.output):]

            states.append(state)
            activations.append(activation)

            action = self.distribution.act_deterministic(*probabilities)
            action_trajectory.append(action)
            observation, reward, done, info = env.step(action)
            observation, reward, done, info = self.preprocessor.modulate((parse_state(observation), reward, done, info),
                                                                      update=False)

            state = observation
            reward_trajectory.append(reward)

            env.render() if render else ""

        return [states, list(zip(*activations)), reward_trajectory, action_trajectory]

    def render_episode(self, env: gym.Env, substeps_per_step=1, act_confidently=True) -> None:
        """Render an episode in the given environment."""
        if not env.render_mode == "human":
            print("WARNING: Rendering is not set to human mode. You will not see any output. Rebuild the environment"
                  "in human render mode.")

        is_recurrent = is_recurrent_model(self.network)
        self.network.reset_states()

        done, step = False, 0

        state, _ = env.reset()
        cumulative_reward = 0
        while not done:
            step += 1

            prepared_state = state.with_leading_dims(time=is_recurrent).dict()
            probabilities = flatten(self.network(prepared_state, training=False))

            if act_confidently:
                action, _ = self.distribution.act_deterministic(*probabilities)
            else:
                action, _ = self.distribution.act(*probabilities)

            observation, reward, terminated, truncated, info = env.step(action)

            cumulative_reward += (reward if "original_reward" not in info else info["original_reward"])
            done = terminated or truncated

            state = observation

        print(f"Finished after {step} steps with a score of {round(cumulative_reward, 4)}. "
              f"{'Good Boy!' if env.spec.reward_threshold is not None and cumulative_reward > env.spec.reward_threshold else ''}")

    def render_episode_jupyter(self, env: TaskWrapper, substeps_per_step=1, act_confidently=True) -> None:
        """Render an episode in the given environment."""
        import mediapy as media

        is_recurrent = is_recurrent_model(self.network)
        self.network.reset_states()

        done, step = False, 0

        frames = []
        state, _ = env.reset()
        cumulative_reward = 0
        while not done:
            frames.append(env.render())
            step += 1

            prepared_state = state.with_leading_dims(time=is_recurrent).dict()
            probabilities = flatten(self.network(prepared_state, training=False))

            if act_confidently:
                action, _ = self.distribution.act_deterministic(*probabilities)
            else:
                action, _ = self.distribution.act(*probabilities)

            observation, reward, terminated, truncated, info = env.step(action)
            cumulative_reward += (reward if "original_reward" not in info else info["original_reward"])
            done = terminated or truncated

            state = observation

        return media.show_video(frames, fps=20)


if __name__ == "__main__":
    print("INVESTIGATING")
    os.chdir("../../../")
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

    # agent_id = 1585500821  # cartpole-v1
    agent_id = 1583256614 # reach task
    agent_007 = PPOAgent.from_agent_state(agent_id, from_iteration="b")

    inv = Investigator.from_agent(agent_007)
    print(inv.list_layer_names())

    # inv.get_activations_over_episode("policy_recurrent_layer", agent_007.env)

    # inv.get_activations_over_episode("policy_recurrent_layer", agent_007.env)

    for i in range(100):
        inv.render_episode(agent_007.env)



import matplotlib.pyplot as plt
import scipy as sp
import sklearn.decomposition as skld
from matplotlib import animation
from mpl_toolkits.mplot3d import Axes3D

def plot_results(self, results, action_data, fixed_point, title: str = "Results", dreiD: bool = False):
    """Plot results of analysis performed by chiefinvestigator

    Args:
        results: Results of analysis to be plotted.
        action_data: Actions selected by agent that are used to color the plots.
        title: Title of the plot. Default is set to "Results"
        :param dreiD:
        :param fixed_point:

    Returns:
        Plot of data to be visualized

    """
    if dreiD is True:
        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        ax.scatter(results[:, 0], results[:, 1], results[:, 2],
                   marker='o', s=5, c=action_data[:])
        if fixed_point is not None:
            ax.scatter(fixed_point[:, 0], fixed_point[:, 1], fixed_point[:, 2],
                       marker='x', s=30)
        plt.title(title)
        ax.set_xlabel('PC1')
        ax.set_ylabel('PC2')
        ax.set_zlabel('PC3')
        plt.show()

    if dreiD is False:
        plt.figure()
        plt.scatter(results[:, 1], results[:, 0],
                    c=action_data[:],
                    label=action_data[:])
        # legend_elements = [Line2D([0], [0], marker='o', label='Action: 0', color='w',
        #                         markerfacecolor='y', markersize=10),
        #                  Line2D([0], [0], marker='o', label='Action: 1', color='w',
        #                        markerfacecolor='tab:purple', markersize=10)]
        # plt.legend(handles=legend_elements)
        plt.title(title)
        plt.xlabel('second component')
        plt.ylabel('first component')
        # plt.pause(0.5)
        plt.show()


def plot_rewards(self, rewards):
    plt.plot(list(range(len(rewards))), rewards)
    plt.xlabel('Number of steps in episode')
    plt.ylabel('Numerical reward')
    plt.title('Reward over episode')
    plt.show()


def timestepwise_pca(self, activation_data, action_data, title: str = "Results"):
    zscores = sp.stats.zscore(activation_data)  # normalization

    pca = skld.PCA(3)
    pca.fit(zscores)
    X_pca = pca.transform(zscores)

    fig = plt.figure()
    plt.xlim([-3, 4])
    plt.ylim([-7.5, 10.5])

    def update(i):
        plot_results(X_pca[i:(i + 10), :], action_data[i:(i + 10), 0], title)  # plot pca results

    anim = animation.FuncAnimation(fig, update, frames=int(len(X_pca) - 1), interval=100)
    anim.save("moving_pca.gif",
              writer='pillow')


import numpy as np
import sklearn.model_selection
import tqdm

from angorapy.analysis.investigators import base_investigator

import tensorflow as tf

from angorapy.analysis.sindy.autoencoder import SindyAutoencoder
from angorapy.analysis.util.sindy import compute_z_derivatives, sindy_library_tf
from angorapy.common.policies import BasePolicyDistribution
from angorapy.tasks.wrappers import TaskWrapper
from angorapy.utilities import core

import pysindy as ps


class Dynamics(base_investigator.Investigator):
    """Investigate dynamics of the system constituted by a recurrent model."""

    def __init__(self, network: tf.keras.Model, distribution: BasePolicyDistribution):
        super().__init__(network, distribution)
        self._data = None
        self.is_prepared = False

        assert self.is_recurrent, "Network is not recurrent, no system to be identified without temporal dependencies."

    def prepare(self, env: TaskWrapper, layer: str, n_states: int):
        """Collect data for the system to be identified from via SINDy."""
        state, info = env.reset(return_info=True)

        activation_collection = []
        for _ in tqdm.tqdm(range(n_states)):
            prepared_state = state.with_leading_dims(time=self.is_recurrent).dict()
            activations = self.get_layer_activations([layer], prepared_state)
            activations[layer] = activations[layer][1]
            activation_collection.append(activations)

            probabilities = util.flatten(activations["output"])
            action, _ = self.distribution.act(*probabilities)

            observation, reward, done, info = env.step(action)

            if done:
                state, info = env.reset(return_info=True)
                self.network.reset_states()
            else:
                state = observation
        self._data = util.stack_dicts(activation_collection)[layer]
        self.train_data, self.val_data = sklearn.model_selection.train_test_split(self._data, shuffle=False, test_size=0.2)
        self.is_prepared = True

    def fit(self):
        """Fit a SINDy regressor to the prepared data."""
        assert self.is_prepared, "Need to prepare data first. Run .prepare(...)."

        model = ps.SINDy(
            optimizer=ps.STLSQ(threshold=0.0001),
            feature_library=ps.PolynomialLibrary(degree=2)
        )

        model.fit(self.train_data, t=0.1)
        model.print()


class LatentDynamics(Dynamics):
    """Discover dynamics in latent variables using the SINDy autoencoder algorithm."""

    def __init__(self, network: tf.keras.Model, distribution: BasePolicyDistribution):
        super().__init__(network, distribution)

        self.loss_weights = {
            "x": 5e-4,
            "y": 5e-5,
            "regularization": 1e-5
        }

    def fit(self, n_epochs: int = 10, batch_size: int = 128):
        data = tf.squeeze(self.train_data)

        sindy_autoencoder = SindyAutoencoder([256, 128, 32], z_dim=32, original_dim=data.shape[-1])
        optimizer = tf.keras.optimizers.Adam()

        dataset = tf.data.Dataset.from_tensor_slices((data, data))
        dataset = dataset.batch(batch_size)

        for epoch in range(n_epochs):

            total_epoch_loss = 0
            for step, (x_train_batch, y_train_batch) in enumerate(dataset):

                with tf.GradientTape() as tape:
                    z_coordinates, reconstruction, dz_prediction = sindy_autoencoder(x_train_batch)

                    dx = np.gradient(x_train_batch, axis=1)  # todo axis=0 or 1?
                    dz = compute_z_derivatives(
                        x_train_batch,
                        dx,
                        weights=[layer.get_weights()[0] for layer in sindy_autoencoder.encoder.layers],
                        biases=[layer.get_weights()[1] for layer in sindy_autoencoder.encoder.layers]
                    )
                    dx_decoded = compute_z_derivatives(
                        z_coordinates,
                        dz_prediction,
                        weights=[layer.get_weights()[0] for layer in sindy_autoencoder.decoder.layers],
                        biases=[layer.get_weights()[1] for layer in sindy_autoencoder.decoder.layers]
                    )

                    reconstruction_loss = tf.reduce_mean((x_train_batch - reconstruction) ** 2)
                    sindy_x_loss = tf.reduce_mean((dx - dx_decoded) ** 2)
                    sindy_z_loss = tf.reduce_mean((dz - dz_prediction) ** 2)
                    sindy_regularization_loss = tf.reduce_mean(tf.abs(sindy_autoencoder.coefficients))

                    loss_value = reconstruction_loss \
                                 + sindy_x_loss * self.loss_weights["x"] \
                                 + sindy_z_loss * self.loss_weights["y"] \
                                 + sindy_regularization_loss * self.loss_weights["regularization"]

                grads = tape.gradient(loss_value, sindy_autoencoder.trainable_weights)
                optimizer.apply_gradients(zip(grads, sindy_autoencoder.trainable_weights))

                total_epoch_loss += float(loss_value)

            print(f"Current loss after {epoch} epochs: {total_epoch_loss}")




#!/usr/bin/env python
"""Functions for gathering experience and communicating it to the main thread."""
import abc
import code
import gc
import random
from typing import Tuple, Any

import numpy as np
import tensorflow as tf
from gymnasium.spaces import Box
from tqdm import tqdm

from angorapy.agent.utils import estimate_episode_advantages
from angorapy.agent.dataio import tf_serialize_example, make_dataset_and_stats, serialize_sample
from angorapy.common.data_buffers import ExperienceBuffer, TimeSequenceExperienceBuffer
from angorapy.common.policies import BasePolicyDistribution
from angorapy.common.senses import Sensation
from angorapy.tasks.wrappers import TaskWrapper
from angorapy.common.const import STORAGE_DIR, DETERMINISTIC
from angorapy.utilities.datatypes import StatBundle
from angorapy.utilities.model_utils import is_recurrent_model
from angorapy.utilities.core import add_state_dims, flatten, env_extract_dims


class BaseGatherer(abc.ABC):

    @abc.abstractmethod
    def collect(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def postprocess(self, buffer: ExperienceBuffer, model: tf.keras.Model, env: TaskWrapper) -> ExperienceBuffer:
        pass

    @abc.abstractmethod
    def select_action(self, *args, **kwargs) -> Tuple[tf.Tensor, tf.Tensor]:
        pass


class Gatherer(BaseGatherer):
    """Standard worker implementation for collecting experience by rolling out a policy.

    This is the default PPO behaviour.
    """

    def __init__(
            self,
            worker_id: int,
            exp_id: int,
            distribution: BasePolicyDistribution,
            horizon: int,
            discount: float,
            lam: float,
            subseq_length: int):
        """

        Args:
            worker_id:
            exp_id:
            horizon:
            lam:
            discount:
            distribution:   policy distribution object
        """
        self.worker_id = worker_id
        self.exp_id = exp_id

        # parameters
        self.distribution = distribution
        self.horizon = horizon
        self.lam = lam
        self.discount = discount
        self.subseq_length = subseq_length

    def collect(self,
                joint: tf.keras.Model,
                env: TaskWrapper,
                collector_id: int) -> StatBundle:
        """Collect a batch shard of experience for a given number of time steps.

        Args:
            joint:          network returning both policy and value
            env:            environment from which to gather the data
            horizon:        the number of steps gathered by this worker
            discount:       discount factor
            lam:            lambda parameter of GAE balancing the tradeoff between bias and variance
            subseq_length:  the length of connected subsequences for TBPTT
            collector_id:   the ID of this gathering, different from the worker's ID
        """
        state: Sensation

        is_recurrent = is_recurrent_model(joint)
        is_continuous = isinstance(env.action_space, Box)
        state_dim, action_dim = env_extract_dims(env)

        if DETERMINISTIC:
            env.seed(1)

        # reset states of potentially recurrent net
        if is_recurrent:
            joint.reset_states()

        # buffer storing the experience and stats
        if is_recurrent:
            assert self.horizon % self.subseq_length == 0, "Subsequence length would require cutting of part of the n_steps."
            buffer = TimeSequenceExperienceBuffer(self.horizon // self.subseq_length, state_dim, action_dim,
                                                  is_continuous, self.subseq_length)
        else:
            buffer = ExperienceBuffer(self.horizon, state_dim, action_dim, is_continuous)

        # go for it
        t, current_episode_return, episode_steps, current_subseq_length = 0, 0, 1, 0
        states, rewards, actions, action_probabilities, values, advantages, dones = [], [], [], [], [], [], []
        episode_endpoints = []
        achieved_goals = []
        state, info = env.reset()

        joint.__call__ = tf.function(joint.__call__)
        with tqdm(total=self.horizon, disable=self.worker_id != 0, desc="Gathering experience...") as pbar:
            while t < self.horizon:
                current_subseq_length += 1

                # based on given state, predict action distribution and state value; need flatten due to tf eager bug
                prepared_state = state.with_leading_dims(time=is_recurrent).dict_as_tf()
                policy_out = flatten(joint(prepared_state, training=False))
                # policy_out = fake_joint_output(joint)

                predicted_distribution_parameters, value = policy_out[:-1], policy_out[-1]
                # from the action distribution sample an action and remember both the action and its probability
                action, action_probability = self.select_action(predicted_distribution_parameters)
                # action, action_probability = env.action_space.sample(), 0.5

                states.append(state)
                values.append(np.squeeze(value))
                actions.append(action)

                action_probabilities.append(action_probability)  # should probably ensure that no probability is ever 0

                # make a step based on the chosen action and collect the reward for this state
                observation, reward, terminated, truncated, info = env.step(
                    np.atleast_1d(action) if is_continuous else action)
                done = terminated or truncated
                current_episode_return += (reward if "original_reward" not in info else info["original_reward"])
                rewards.append(reward)
                dones.append(done)

                if hasattr(info, "keys") and "achieved_goal" in info.keys():
                    achieved_goals.append(info["achieved_goal"])

                # if recurrent, at a subsequence breakpoint/episode end stack the n_steps and buffer them
                if is_recurrent and (current_subseq_length == self.subseq_length or done):
                    buffer.push_seq_to_buffer(states=states,
                                              actions=actions,
                                              action_probabilities=action_probabilities,
                                              values=values[-current_subseq_length:],
                                              episode_ended=done)

                    # clear the buffered information
                    states, actions, action_probabilities = [], [], []
                    current_subseq_length = 0

                # depending on whether the state is terminal, choose the next state
                if done:
                    episode_endpoints.append(t)

                    # calculate advantages for the finished episode, where the last value is 0 since it refers to the
                    # terminal state that we just observed
                    episode_advantages = estimate_episode_advantages(rewards[-episode_steps:],
                                                                     values[-episode_steps:] + [0],
                                                                     self.discount, self.lam)
                    episode_returns = episode_advantages + values[-episode_steps:]

                    if is_recurrent:
                        # skip as many steps as are missing to fill the subsequence, then push adv ant ret to buffer
                        skip_steps = self.subseq_length - (t % self.subseq_length) - 1
                        t += skip_steps
                        pbar.update(skip_steps)

                        buffer.push_adv_ret_to_buffer(episode_advantages, episode_returns)
                    else:
                        advantages.append(episode_advantages)

                    # update/reset some statistics and trackers
                    buffer.episode_lengths.append(episode_steps)
                    buffer.episode_rewards.append(current_episode_return)
                    buffer.episodes_completed += 1
                    episode_steps = 1
                    current_episode_return = 0

                    if "auxiliary_performances" in info.keys():
                        for key, value in info["auxiliary_performances"].items():
                            if key not in buffer.auxiliary_performances.keys():
                                buffer.auxiliary_performances[key] = []

                            buffer.auxiliary_performances[key].append(value)

                    # reset environment to receive next episodes initial state
                    state, info = env.reset()

                    if is_recurrent:
                        joint.reset_states()
                else:
                    state = observation
                    episode_steps += 1

                t += 1
                pbar.update(1)

        # add auxiliary performance to buffer even if no episode is finished
        if "auxiliary_performances" in info.keys():
            for key, value in info["auxiliary_performances"].items():
                if key not in buffer.auxiliary_performances.keys():
                    buffer.auxiliary_performances[key] = []

        # get last non-visited state value to incorporate it into the advantage estimation of last visited state
        values.append(
            np.squeeze(joint(add_state_dims(state, dims=2 if is_recurrent else 1).dict(), training=False)[-1]))

        # if there was at least one step in the environment after the last episode end, calculate advantages for them
        if episode_steps > 1:
            leftover_advantages = estimate_episode_advantages(rewards[-episode_steps + 1:], values[-episode_steps:],
                                                              self.discount, self.lam)
            if is_recurrent:
                leftover_returns = leftover_advantages + values[-len(leftover_advantages) - 1:-1]
                buffer.push_adv_ret_to_buffer(leftover_advantages, leftover_returns)
            else:
                advantages.append(leftover_advantages)

        # if not recurrent, fill the buffer with everything we gathered
        if not is_recurrent:
            values = np.array(values, dtype="float32")

            # write to the buffer
            advantages = np.hstack(advantages).astype("float32")
            returns = advantages + values[:-1]
            buffer.fill(states,
                        np.array(actions, dtype="float32" if is_continuous else "int32"),
                        np.array(action_probabilities, dtype="float32"),
                        advantages,
                        returns,
                        values[:-1],
                        np.array(dones),
                        np.array(achieved_goals))

        # postprocessing steps
        buffer = self.postprocess(buffer, joint, env)

        # convert buffer to dataset and save it to tf record
        dataset, stats = make_dataset_and_stats(buffer)
        with tf.io.TFRecordWriter(f"{STORAGE_DIR}/{self.exp_id}_data_{collector_id}.tfrecord") as file_writer:
            feature_names = ([sense for sense in Sensation.sense_names if sense in observation]
                             + ["action", "action_prob", "return", "advantage", "value", "done", "mask"])

            for batch in dataset:
                inputs = [batch[f] for f in feature_names]
                record = serialize_sample(*inputs, feature_names=feature_names)

                file_writer.write(record)

        gc.collect()
        tf.keras.backend.clear_session()
        tf.compat.v1.reset_default_graph()

        return stats

    def postprocess(self, buffer: ExperienceBuffer, model: tf.keras.Model, env: TaskWrapper):
        """Postprocess the gathered data."""
        buffer.normalize_advantages()

        return buffer

    def select_action(self, predicted_parameters: list) -> np.ndarray:
        """Standard action selection where an action is sampled fully from the predicted distribution."""
        action, action_probability = self.distribution.act_numpy(*predicted_parameters)
        action = action if not DETERMINISTIC else np.zeros_like(action)

        return action, action_probability


class EpsilonGreedyGatherer(Gatherer):
    """Exemplary epsilon greedy gathering strategy.

    This is not safe! At some point, the probability of a uniform sample may become zero under the current policy and
    thereby the log probability goes to negative infinity, leading to the optimization to collapse.
    """

    def select_action(self, predicted_parameters: list) -> Tuple[tf.Tensor, np.ndarray]:
        if random.random() < 0.9:
            action, action_probability = super(EpsilonGreedyGatherer, self).select_action(predicted_parameters)
        else:
            action = tf.cast(self.distribution.action_space.sample(), tf.float32)
            action_probability = self.distribution.log_probability(action, *predicted_parameters)

        return action, action_probability


def evaluate(policy: tf.keras.Model, env: TaskWrapper, distribution: BasePolicyDistribution,
             act_confidently=False) -> Tuple[int, int, Any, dict]:
    """Evaluate one episode of the given environment following the given policy."""
    policy.reset_states()
    is_recurrent = is_recurrent_model(policy)
    is_continuous = isinstance(env.action_space, Box)

    done = False
    state, info = env.reset()
    cumulative_reward = 0
    steps = 0
    while not done:
        prepared_state = state.with_leading_dims(time=is_recurrent).dict_as_tf()
        probabilities = flatten(policy(prepared_state))

        if not act_confidently:
            action, _ = distribution.act(*probabilities)
        else:
            action, _ = distribution.act_deterministic(*probabilities)

        observation, reward, terminated, truncated, info = env.step(np.atleast_1d(action) if is_continuous else action)
        cumulative_reward += info["original_reward"]
        observation = observation
        done = terminated or truncated
        state = observation
        steps += 1

    auxiliary_performances = {}
    if "auxiliary_performances" in info.keys():
        for key, value in info["auxiliary_performances"].items():
            if key not in auxiliary_performances.keys():
                auxiliary_performances[key] = []

            auxiliary_performances[key].append(value)

    eps_class = env.unwrapped.current_target_finger if hasattr(env.unwrapped, "current_target_finger") else None

    return steps, cumulative_reward, eps_class, auxiliary_performances


def fake_env_step(env: TaskWrapper):
    """Return a random step imitating the given environment without actually stepping the environment."""
    return env.observation(env.observation_space.sample()), 1, False, {}


def fake_joint_output(joint):
    """Return a random output of a network without calling the network."""
    outshape = flatten(joint.output_shape)
    for i in range(len(outshape)):
        outshape[i] = [d if d is not None else 1 for d in outshape[i]]

    return np.float32(np.random.random(outshape[0])), \
        np.float32(np.random.random(outshape[1])), \
        np.float32(np.random.random(outshape[2]))



import unittest
from typing import Sequence, Union

import numpy as np
import pytest

import angorapy as ap

from dm_control import mjcf


class _TestTask(ap.tasks.AnthropomorphicEnv):
    def _set_default_reward_function_and_config(self):
        self.reward_function = lambda x, info: 0
        self.reward_config = {}

    def assert_reward_setup(self):
        pass

    def _env_setup(self, initial_state):
        pass

    def _sample_goal(self):
        return np.zeros(0)


class _TestRobot(ap.tasks.world_building.Robot):

    @property
    def joints(self) -> Sequence[mjcf.Element]:
        return []

    @property
    def actuators(self) -> Sequence[mjcf.Element]:
        return []

    def _parse_entity(self) -> None:
        pass

    def _setup_entity(self) -> None:
        pass

    @property
    def root_body(self) -> Union[None, mjcf.Element]:
        return None


def test_minimal_env():
    robot = _TestRobot(mjcf.RootElement())
    task = _TestTask(robot)

    state = task.reset()
    for _ in range(100):
        state, r, dterm, dtrunc, info = task.step(task.action_space.sample())
        print(state)

    assert True



if __name__ == '__main__':
    unittest.main()



from angorapy import agent
from angorapy import common
from angorapy import models
from angorapy import tasks
from angorapy import utilities
from angorapy import analysis

from angorapy.tasks.registration import make_task
from angorapy.tasks.registration import make_task as make_env  # For backwards compatibility

from angorapy.agent import Agent

# Builtin models
from angorapy.models import simple
from angorapy.models import shadow
from angorapy.models import shadow_v2



from angorapy.tasks.world_building.base import PhysicalWorld
from angorapy.tasks.world_building.entities import Stage
from angorapy.tasks.envs.dexterity.mujoco_model.robot import ShadowHand, ShadowHandReach


class ShadowHandReachWorld(PhysicalWorld):

    def __init__(self):
        self._stage = Stage()
        self._robot = ShadowHandReach()

        self._stage.attach(self._robot)

    @property
    def stage(self) -> Stage:
        return self._stage

    @property
    def robot(self) -> ShadowHand:
        return self._robot


if __name__ == '__main__':
    world = ShadowHandReach()
    print(world.root.to_xml_string())



import json
import os
import re
import shutil
import sys
from json import JSONDecodeError

from bokeh import embed
from tqdm import tqdm

from angorapy.models import MODELS_AVAILABLE

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from angorapy.utilities.monitor.statistical_plots import plot_episode_box_plots, plot_per_receptor_mean

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

import flask
from flask import request, Blueprint, send_from_directory
from flask_jsglue import JSGlue

from angorapy.agent.ppo_agent import PPOAgent
from angorapy.common.const import PATH_TO_EXPERIMENTS, BASE_SAVE_PATH
from angorapy.utilities.monitor.training_plots import plot_memory_usage, plot_execution_times, plot_preprocessor, \
    plot_reward_progress, plot_loss, plot_length_progress, plot_distribution, compare_reward_progress, \
    grouped_reward_progress, group_preview, plot_aux_perf_progress
from angorapy.utilities.statistics import ignore_none

from angorapy.monitor import app


agents = Blueprint("agents", __name__, static_folder="storage/saved_models/states")
exps = Blueprint("exps", __name__, static_folder="storage/experiments")

app.register_blueprint(agents)
app.register_blueprint(exps)

jsglue = JSGlue(app)


@app.route("/")
def overview():
    """Write Overview page."""
    exp_dir = PATH_TO_EXPERIMENTS
    experiment_paths = [os.path.join(exp_dir, p) for p in os.listdir(exp_dir)]

    experiments = {}
    envs_available = set()
    for exp_path in tqdm(experiment_paths):

        eid_m = re.match("[0-9]+", str(exp_path.split("/")[-1]))
        if eid_m:
            eid = eid_m.group(0)
            model_path = os.path.join(BASE_SAVE_PATH, eid)

            if os.path.isfile(os.path.join(exp_path, "progress.json")):
                try:
                    with open(os.path.join(exp_path, "progress.json"), "r") as f:
                        progress = json.load(f)
                except:
                    continue

                try:
                    with open(os.path.join(exp_path, "meta.json"), "r") as f:
                        meta = json.load(f)
                except:
                    pass

                agent_parameters = {}
                model_available = False
                if os.path.isfile(os.path.join(model_path, "best/weights.index")):
                    model_available = True

                    # with open(os.path.join(model_path, "best/parameters.json")) as f:
                    #     agent_parameters = json.load(f)

                reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
                    meta["environment"]["reward_threshold"])
                iterations = len(progress["rewards"]["mean"])
                max_performance = ignore_none(max, progress["rewards"]["mean"])
                is_success = False
                if iterations >= 0 and reward_threshold is None:
                    is_success = "maybe"
                elif max_performance is not None and max_performance > reward_threshold:
                    is_success = True

                architecture = "Any" if "architecture" not in meta["hyperparameters"] else meta["hyperparameters"]["architecture"]
                model = "Any" if "model" not in meta["hyperparameters"] else meta["hyperparameters"]["model"]
                if model_available:
                    architecture = agent_parameters.get("builder_function_name", architecture)
                    for mname in MODELS_AVAILABLE:
                        if mname in architecture:
                            architecture = mname

                experiments.update({
                    eid: {
                        "env": meta["environment"]["name"],
                        "date": meta["date"],
                        "host": meta["host"] if "host" in meta else "unknown",
                        "iterations": iterations,
                        "max_reward": max_performance if iterations > 0 else None,
                        "is_success": is_success,
                        "bookmark": meta["bookmark"] if "bookmark" in meta else False,
                        "config_name": meta["config"] if "config" in meta else "unknown",
                        "reward": meta["reward_function"] if "reward_function" in meta else "None",
                        "model": architecture + "[" + model + "]",
                        "model_available": model_available,
                        "gatherer": meta["hyperparameters"].get("gatherer", "default"),
                        "policy": meta["hyperparameters"].get("distribution", "default")
                    }
                })

                envs_available.add(meta["environment"]["name"])

    return flask.render_template("overview.html", exps=experiments, envs_available=envs_available)


@app.route("/groups")
def view_groups():
    """Build group view page."""
    exp_dir = PATH_TO_EXPERIMENTS
    experiment_paths = [os.path.join(exp_dir, p) for p in os.listdir(exp_dir)]

    experiments_by_groups = {}
    envs_available = set()
    for exp_path in experiment_paths:

        eid_m = re.match("[0-9]+", str(exp_path.split("/")[-1]))
        if eid_m:
            eid = eid_m.group(0)
            model_path = os.path.join(BASE_SAVE_PATH, eid)

            if os.path.isfile(os.path.join(exp_path, "progress.json")):
                with open(os.path.join(exp_path, "progress.json"), "r") as f:
                    progress = json.load(f)

                with open(os.path.join(exp_path, "meta.json"), "r") as f:
                    try:
                        meta = json.load(f)
                    except JSONDecodeError as jse:
                        continue

                exp_group = meta.get("experiment_group", "n/a")

                if not exp_group in experiments_by_groups.keys():
                    experiments_by_groups[exp_group] = {}

                model_available = False
                if os.path.isfile(os.path.join(model_path, "best/weights.index")):
                    model_available = True

                reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
                    meta["environment"]["reward_threshold"])
                iterations = len(progress["rewards"]["mean"])
                max_performance = ignore_none(max, progress["rewards"]["mean"])
                is_success = False
                if iterations >= 0 and reward_threshold is None:
                    is_success = "maybe"
                elif max_performance is not None and max_performance > reward_threshold:
                    is_success = True

                architecture = "Any" if "architecture" not in meta["hyperparameters"] else meta["hyperparameters"]["architecture"]
                model = "Any" if "model" not in meta["hyperparameters"] else meta["hyperparameters"]["model"]

                experiments_by_groups[exp_group].update({
                    eid: {
                        "env": meta["environment"]["name"],
                        "date": meta["date"],
                        "host": meta["host"] if "host" in meta else "unknown",
                        "iterations": iterations,
                        "max_reward": max_performance if iterations > 0 else "N/A",
                        "is_success": is_success,
                        "bookmark": meta["bookmark"] if "bookmark" in meta else False,
                        "config_name": meta["config"] if "config" in meta else "unknown",
                        "reward": meta["reward_function"] if "reward_function" in meta else "None",
                        "model": architecture + "[" + model + "]",
                        "model_available": model_available
                    }
                })

                envs_available.add(meta["environment"]["name"])

    return flask.render_template("groups.html", exps=experiments_by_groups, envs_available=envs_available)


@app.route("/group/", methods=("POST", "GET"))
def show_group():
    """Analyze the summarized results of a group of experiments."""
    ids = request.args.getlist('ids')
    group_name = request.args.get('group')
    info = {
        "ids": ids,
        "group_name": group_name,
        "plots": {}
    }

    progress_reports = {}
    metas = {}
    environments = []

    for exp_id in ids:
        path = f"{PATH_TO_EXPERIMENTS}/{exp_id}"
        with open(os.path.join(path, "progress.json"), "r") as f:
            progress = json.load(f)

        with open(os.path.join(path, "meta.json"), "r") as f:
            meta = json.load(f)

        progress_reports.update({f"{exp_id}": progress})
        metas[str(exp_id)] = meta

        environments.append(meta["environment"]["name"])

    environments = set(environments)
    environment_reward_threshold = None
    if len(environments) == 1:
        environment_reward_threshold = meta["environment"].get("reward_threshold", None)  # take last exps threshold

    info["plots"]["reward_grouped"] = grouped_reward_progress(
        {id: (group_name, progress["rewards"]) for id, progress in progress_reports.items()},
        environment_reward_threshold
    )

    return flask.render_template("group_analysis.html", info=info)


@app.route("/make_group_preview", methods=("POST", "GET"))
def make_group_preview():
    """Analyze the summarized results of a group of experiments."""
    group_names = request.json['names']
    group_names = group_names.split(",")

    exp_dir = PATH_TO_EXPERIMENTS
    experiment_paths = [os.path.join(exp_dir, p) for p in os.listdir(exp_dir)]

    environments = {}
    reward_thresholds = {}
    experiments_by_groups = {}
    envs_available = set()
    for exp_path in experiment_paths:

        eid_m = re.match("[0-9]+", str(exp_path.split("/")[-1]))
        if eid_m:
            eid = eid_m.group(0)
            model_path = os.path.join(BASE_SAVE_PATH, eid)

            if os.path.isfile(os.path.join(exp_path, "progress.json")):
                with open(os.path.join(exp_path, "progress.json"), "r") as f:
                    progress = json.load(f)

                with open(os.path.join(exp_path, "meta.json"), "r") as f:
                    try:
                        meta = json.load(f)
                    except JSONDecodeError as jserr:
                        continue

                exp_group = meta.get("experiment_group", "n/a")

                if exp_group not in group_names:
                    continue

                reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(meta["environment"]["reward_threshold"])

                if not exp_group in experiments_by_groups.keys():
                    experiments_by_groups[exp_group] = {}
                    reward_thresholds[exp_group] = reward_threshold
                    environments[exp_group] = meta["environment"]["name"]

                envs_available.add(meta["environment"]["name"])

                experiments_by_groups[exp_group].update({
                    eid: progress
                })

    return flask.render_template("update_group_preview.html",
                                 plot=group_preview(experiments_by_groups,
                                                    alt_titles=environments,
                                                    reward_thresholds=reward_thresholds))


@app.route("/benchmarks")
def benchmarks():
    """Write Benchmark page."""
    exp_dir = PATH_TO_EXPERIMENTS
    experiment_paths = [os.path.join(exp_dir, p) for p in os.listdir(exp_dir)]

    experiments = {}
    envs_available = set()
    for path in experiment_paths:
        eid_m = re.match("[0-9]+", str(path.split("/")[-1]))
        if eid_m:
            eid = eid_m.group(0)

            if os.path.isfile(os.path.join(path, "progress.json")):
                with open(os.path.join(path, "progress.json"), "r") as f:
                    progress = json.load(f)

                with open(os.path.join(path, "meta.json"), "r") as f:
                    meta = json.load(f)

                # for vn, vector in progress.items():
                #     replace()

                reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
                    meta["environment"]["reward_threshold"])
                iterations = len(progress["rewards"]["mean"])
                experiments.update({
                    eid: {
                        "env": meta["environment"]["name"],
                        "date": meta["date"],
                        "host": meta["host"] if "host" in meta else "unknown",
                        "iterations": iterations,
                        "max_reward": max(progress["rewards"]["mean"]) if iterations > 0 else "N/A",
                        "is_success": False if iterations == 0 else ("maybe" if reward_threshold is None else max(
                            progress["rewards"]["mean"]) > reward_threshold),
                        "bookmark": meta["bookmark"] if "bookmark" in meta else False
                    }
                })

                envs_available.add(meta["environment"]["name"])

    return flask.render_template("overview.html", exps=experiments, envs_available=envs_available)


@app.route("/bookmark", methods=("POST", "GET"))
def bookmark():
    """Bookmark an experiment."""
    if request.method == "POST":
        try:
            exp_id = int(request.json['id'])
            path = f"{PATH_TO_EXPERIMENTS}{exp_id}"

            with open(os.path.join(path, "meta.json"), "r") as f:
                meta = json.load(f)
                current_status = "bookmark" in meta and meta["bookmark"]

            meta.update({"bookmark": not current_status})

            with open(os.path.join(path, "meta.json"), "w") as f:
                json.dump(meta, f)

        except Exception as e:
            return {"success": e.__repr__()}
    else:
        return {"success": "no post"}

    return {"success": "success"}


@app.route("/experiment/<int:exp_id>", methods=("POST", "GET"))
def show_experiment(exp_id):
    """Show experiment of given ID."""
    experiment_paths = sorted([int(p) for p in os.listdir(f"{PATH_TO_EXPERIMENTS}") if re.match("^[0-9]*$", p)])
    current_index = experiment_paths.index(exp_id)

    path = f"{PATH_TO_EXPERIMENTS}/{exp_id}"
    with open(os.path.join(path, "progress.json"), "r") as f:
        progress = json.load(f)

    with open(os.path.join(path, "meta.json"), "r") as f:
        meta = json.load(f)

    info = dict(
        env=meta["environment"]["name"],
        config=meta["config"] if "config" in meta else None,
        host=meta["host"] if "host" in meta else "unknown",
        current_id=exp_id,
        next_id=experiment_paths[current_index + 1] if current_index != len(experiment_paths) - 1 else None,
        prev_id=experiment_paths[current_index - 1] if current_index != 0 else None,
        hps=meta["hyperparameters"],
        env_meta=meta["environment"],
        reward_function=meta["reward_function"] if "reward_function" in meta else {},
        iterations=meta["iterations"] if "iterations" in meta else None
    )

    plots = {"normalization": {}}

    stats_path = os.path.join(path, "statistics.json")
    if os.path.isfile(stats_path):
        with open(stats_path, "r") as f:
            stats = json.load(f)

        info.update(dict(
            statistics=stats
        ))

        if "used_memory" in stats:
            plots["mem_usage"] = embed.components(plot_memory_usage(stats["used_memory"]))

        if "cycle_timings" in stats:
            plots["timings"] = embed.components(plot_execution_times(stats["cycle_timings"],
                                                    stats.get("optimization_timings"),
                                                    stats.get("gathering_timings")))

    if "RewardNormalizationTransformer" in progress["preprocessors"]:
        plots["normalization"]["reward"] = embed.components(plot_preprocessor(progress["preprocessors"]["RewardNormalizationTransformer"]))

    if "StateNormalizationTransformer" in progress["preprocessors"]:
        plots["normalization"]["state"] = embed.components(plot_preprocessor(progress["preprocessors"]["StateNormalizationTransformer"]))

    reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
        meta["environment"]["reward_threshold"])

    plots["reward_distribution"] = embed.components(plot_distribution(progress["rewards"]["last_cycle"], "Rewards (Last Cycle)", color=0))
    plots["length_distribution"] = embed.components(plot_distribution(progress["lengths"]["last_cycle"], "Episode Lengths (Last Cycle)", color=1))
    cycles_loaded = []
    if "loaded_at" in stats.keys():
        cycles_loaded = stats["loaded_at"]
    plots["reward_progress"] = embed.components(plot_reward_progress(progress["rewards"], cycles_loaded, reward_threshold=reward_threshold))
    plots["length_progress"] = embed.components(plot_length_progress(progress["lengths"], cycles_loaded))
    plots["policy_loss"] = embed.components(plot_loss(progress["ploss"], progress["rewards"]["mean"], "Policy Loss", color_id=0))
    plots["value_loss"] = embed.components(plot_loss(progress["vloss"], progress["rewards"]["mean"], "Value Loss", color_id=1))
    plots["entropies"] = embed.components(plot_loss(progress["entropies"], progress["rewards"]["mean"], "Entropy", color_id=2))

    if "per_receptor_mean" in stats.keys():
        plots["per_receptor_mean"] = embed.components(plot_per_receptor_mean(stats["per_receptor_mean"]))

    plots["auxiliary_plots"] = {}
    if "auxiliary_performances" in stats.keys():
        for aux_perf_key, aux_perf_values in stats["auxiliary_performances"].items():
            plots["auxiliary_plots"][aux_perf_key] = embed.components(
                plot_aux_perf_progress(aux_perf_values, cycles_loaded, perf_name=aux_perf_key))

    info.update(dict(
        plots=plots
    ))

    return flask.render_template("experiment.html", info=info)


@app.route("/compare/", methods=("POST", "GET"))
def show_comparison():
    """Compare experiments of given IDs."""
    ids = request.args.getlist('ids')

    info = {
        "ids": ids,
        "plots": {}
    }

    progress_reports = {}
    metas = {}
    for exp_id in ids:
        path = f"{PATH_TO_EXPERIMENTS}/{exp_id}"
        with open(os.path.join(path, "progress.json"), "r") as f:
            progress = json.load(f)

        with open(os.path.join(path, "meta.json"), "r") as f:
            try:
                meta = json.load(f)
            except:
                continue

        progress_reports.update({f"{exp_id}": progress})
        metas[str(exp_id)] = meta

    info["plots"]["reward"] = embed.components(compare_reward_progress(
        {id: progress["rewards"] for id, progress in progress_reports.items()},
        None
    ))

    info["plots"]["reward_grouped"] = embed.components(grouped_reward_progress(
        {id: (metas[str(id)]["hyperparameters"]["epochs_per_cycle"], progress["rewards"]) for id, progress in progress_reports.items() if metas[str(id)]["hyperparameters"]["epochs_per_cycle"] > 30},
        None
    ))

    return flask.render_template("compare.html", info=info)


@app.route("/expfile/<int:exp_id>/<path:filename>")
def expfile(exp_id, filename):
    path = os.path.abspath(os.path.join(PATH_TO_EXPERIMENTS, str(exp_id)))
    return send_from_directory(path, filename, as_attachment=True)


@app.route("/_clear_all_empty")
def clear_all_empty():
    """Delete all experiments stored that have less than 2 episodes finished."""
    experiment_paths = [os.path.join(PATH_TO_EXPERIMENTS, p) for p in os.listdir(PATH_TO_EXPERIMENTS)]

    deleted = 0
    for path in experiment_paths:
        if re.match("[0-9]+", str(path.split("/")[-1])):
            if os.path.isfile(os.path.join(path, "progress.json")):
                try:
                    with open(os.path.join(path, "progress.json"), "r") as f:
                        progress = json.load(f)
                except Exception:
                    # delete corrupted
                    shutil.rmtree(path)
                    deleted += 1
                    continue

                if "rewards" in progress and len(progress["rewards"]["mean"]) < 2:
                    shutil.rmtree(path)
                    deleted += 1

    return {"deleted": deleted}


@app.route("/_clear_all_short")
def clear_all_short():
    """Delete all experiments stored that have less than 10 cycles finished."""
    experiment_paths = [os.path.join(PATH_TO_EXPERIMENTS, p) for p in os.listdir(PATH_TO_EXPERIMENTS)]

    deleted = 0
    for path in experiment_paths:
        if re.match("[0-9]+", str(path.split("/")[-1])):
            if os.path.isfile(os.path.join(path, "progress.json")):
                try:
                    with open(os.path.join(path, "progress.json"), "r") as f:
                        progress = json.load(f)
                except Exception:
                    # delete corrupted
                    shutil.rmtree(path)
                    deleted += 1
                    continue

                if "rewards" in progress and len(progress["rewards"]["mean"]) <= 30:
                    shutil.rmtree(path)
                    deleted += 1

    return {"deleted": deleted}


@app.route('/delete_experiment', methods=['GET', 'POST'])
def delete_experiment():
    """Delete an experiment of posted id."""
    if request.method == "POST":
        try:
            eid = request.json['id']
            eid = int(eid)  # make sure its a number
            shutil.rmtree(os.path.join(PATH_TO_EXPERIMENTS, str(eid)))
        except Exception:
            return {"success": False}
    else:
        return {"success": False}

    return {"success": True}


@app.route('/delete_experiments', methods=['GET', 'POST'])
def delete_experiments():
    """Delete multiple experiments of posted ids."""
    if request.method == "POST":
        try:
            ids = request.json['ids']
            ids = ids.split(",")
            ids = list(map(int, ids))

            for id in ids:
                shutil.rmtree(os.path.join(PATH_TO_EXPERIMENTS, str(id)))
        except Exception:
            return {"success": False}
    else:
        return {"success": False}

    return {"success": True}


@app.route('/regroup_experiments', methods=['GET', 'POST'])
def regroup_experiments():
    """Assign new group to multiple experiments of posted ids."""
    if request.method == "POST":
        try:
            ids = request.json['ids']
            ids = ids.split(",")
            ids = list(map(int, ids))
            new_group = request.json['group']

            for id in ids:
                with open(os.path.join(PATH_TO_EXPERIMENTS, str(id), "meta.json"), "r") as f:
                    data = json.load(f)

                data.update({"experiment_group": new_group})

                with open(os.path.join(PATH_TO_EXPERIMENTS, str(id), "meta.json"), "w") as f:
                    json.dump(data, f)

        except Exception:
            return {"success": False}
    else:
        return {"success": False}

    return {"success": True}


@app.route("/evaluate", methods=("POST", "GET"))
def evaluate():
    """Evaluate an agent."""
    if request.method == "POST":
        try:
            agent = PPOAgent.from_agent_state(request.json['id'])
            evaluation_stats, _ = agent.evaluate(10, save=True)

            return {"results": evaluation_stats._asdict()}

        except Exception as e:
            return {"success": e.__repr__()}

    return {"success": "success"}



#!/usr/bin/env python
"""Pretrain the visual component."""
import os
import sys

import mujoco

from angorapy.common.loss import PoseEstimationLoss

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from angorapy import make_task
from angorapy.tasks.core import AnthropomorphicEnv

from angorapy.models.convolutional import OpenAIEncoder

import argparse
import math

import numpy as np
import tensorflow_datasets as tfds
from tensorflow.python.data import AUTOTUNE

# os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from typing import Tuple

import argcomplete
import tensorflow as tf

from angorapy.common.const import PRETRAINED_COMPONENTS_PATH, VISION_WH
from angorapy.utilities.data_generation import gen_cube_quats_prediction_data, load_dataset, load_unrendered_dataset

import tensorflow_graphics.geometry.transformation as tfg

from mpi4py import MPI


mpi_rank = MPI.COMM_WORLD.Get_rank()
is_root = mpi_rank == 0

# deallocate the GPU from all but the root process
if not is_root:
    tf.config.set_visible_devices([], 'GPU')

# Prevent "Failed to find dnn implementation" error
gpus = tf.config.list_physical_devices("GPU")
if gpus:
    tf.config.experimental.set_memory_growth(gpus[0], True)


@tf.function
def rotational_diff_metric(y_true, y_pred):
    rot_true = y_true[..., 3:]
    rot_pred = y_pred[..., 3:]

    return tfg.quaternion.relative_angle(rot_true, rot_pred) * tf.constant(180. / math.pi)


@tf.function
def positional_diff_metric(y_true, y_pred):
    """Gives positional difference in millimeters."""
    pos_true = y_true[..., :3]
    pos_pred = y_pred[..., :3]

    return tf.linalg.norm(pos_true - pos_pred, axis=-1) * 1000


hand_env = make_task("ManipulateBlockDiscreteAsynchronous-v0",
                     render_mode="rgb_array")
hand_env.reset()
# hand_env.render()

env_unwrapped: AnthropomorphicEnv = hand_env.unwrapped
model = env_unwrapped.model
data = env_unwrapped.data
renderer = mujoco.Renderer(model, height=VISION_WH, width=VISION_WH)

cameras = [env_unwrapped._get_viewer("rgb_array").cam]
cameras[-1].distance = 0.3  # zoom in

for i in range(1, 3):
    cameras.append(mujoco.MjvCamera())
    cameras[-1].type = mujoco.mjtCamera.mjCAMERA_FREE
    cameras[-1].fixedcamid = -1
    for j in range(3):
        cameras[-1].lookat[j] = np.median(np.copy(data.geom_xpos[:, j]))
    cameras[-1].distance = np.copy(model.stat.extent)

    cameras[-1].distance = 0.3  # zoom in
    cameras[-1].azimuth = 0.0 + [20, -20][i - 1]  # wrist to the bottom
    cameras[-1].elevation = -90.0 + [20, -20][i - 1]  # top down view
    cameras[-1].lookat[0] = 0.35  # slightly move forward

rotation_quats = (
    (tfg.quaternion.from_axis_angle(np.array([[1., 0., 0.]]), np.array([[-1 * np.pi / 2]])),
     tfg.quaternion.from_axis_angle(np.array([[1., 0., 0.]]), np.array([[1 * np.pi / 2]]))),
    (tfg.quaternion.from_axis_angle(np.array([[0., 1., 0.]]), np.array([[-1 * np.pi / 2]])),
     tfg.quaternion.from_axis_angle(np.array([[0., 1., 0.]]), np.array([[1 * np.pi / 2]]))),
    (tfg.quaternion.from_axis_angle(np.array([[0., 0., 1.]]), np.array([[-1 * np.pi / 2]])),
     tfg.quaternion.from_axis_angle(np.array([[0., 0., 1.]]), np.array([[1 * np.pi / 2]])))
)


def render(sim_state):
    hand_env.set_state(sim_state[:model.nq], sim_state[model.nq:])

    # Choose a random number between 0 and 1
    p = np.random.random()

    # Get the quaternion for the current orientation of the body
    object = data.jnt("block/object:joint/")
    quat = data.jnt("block/object:joint/").qpos[3:]
    original_qpos = object.qpos.copy()

    if p < 0.2:  # Leave the object pose as is
        pass
    elif p < .6:  # Rotate the object by 90 degrees around its main axes
        # define the rotation
        axis = np.random.choice([0, 1, 2])
        sign = np.random.choice([0, 1])

        # rotate the body by 90 degrees around the chosen axis
        rotation_quat = rotation_quats[axis][sign]

        # Apply the rotation to the body quaternion
        new_quat = tfg.quaternion.multiply(quat, rotation_quat)
        object.qpos[3:] = new_quat
    else:  # "Jitter" the object by adding Gaussian noise to position and orientation
        # Add Gaussian noise to the object position
        noise_pos = np.random.normal(loc=0, scale=0.01, size=3)
        object.qpos[:3] += noise_pos

        # Add Gaussian noise to the object orientation
        noise_quat = np.random.normal(loc=0, scale=0.01, size=4)
        noise_quat /= np.linalg.norm(noise_quat)
        new_quat = tfg.quaternion.multiply(quat, noise_quat)
        object.qpos[3:] = new_quat

    mujoco.mj_forward(model, data)

    images = []
    for cam in cameras:
        renderer.update_scene(data, camera=cam)
        images.append(renderer.render().copy())

    qpos = object.qpos.copy()
    object.qpos = original_qpos

    images = tf.cast(tf.concat(images, axis=-1) / 255, dtype=tf.float32)
    qpos = tf.cast(qpos, dtype=tf.float32)

    # print(tf.reduce_max(images), tf.reduce_mean(images), tf.reduce_min(images))

    # plot slices of 3 channels in images next to each other
    # import matplotlib
    # matplotlib.use('TkAgg')
    # import matplotlib.pyplot as plt
    # fig, axs = plt.subplots(1, len(cameras))
    # for i in range(len(cameras)):
    #     ax = axs[i] if len(cameras) > 1 else axs
    #     ax.imshow(images[:, :, 3*i:3*(i+1)])
    # plt.show()
    # exit()

    return images, qpos


def pretrain_on_object_pose(pretrainable_component: tf.keras.Model,
                            epochs: int,
                            batch_size: int,
                            n_samples: int,
                            n_cameras=1,
                            load_data: bool = False,
                            name="visual_op",
                            load_from: str = None):
    """Pretrain a visual component on prediction of cube position."""
    data_name = f"storage/data/pretraining/pose_data_{n_samples}"
    data_path = f"{data_name}_{mpi_rank}.tfrecord"
    if not load_data:
        gen_cube_quats_prediction_data(
            n_samples,
            data_path,
        )

    if is_root:
        # get all filenames starting with data_name from the data directory
        filenames = tf.io.gfile.glob(f"{data_name}*.tfrecord")

        if len(filenames) == 0:
            raise FileNotFoundError(f"Could not find any files matching {data_name}*.tfrecord")

        dataset = load_unrendered_dataset(filenames)

        dataset = dataset.map(lambda x, y: tf.py_function(func=render, inp=[x], Tout=[tf.float32, tf.float32]))
        # dataset = dataset.map(lambda x, y: render(x))

        n_testset = 512
        n_valset = 256

        testset = dataset.take(n_testset)
        trainset = dataset.skip(n_testset)
        valset, trainset = trainset.take(n_valset), trainset.skip(n_valset)

        trainset = trainset.batch(batch_size, drop_remainder=True)
        valset = valset.batch(batch_size, drop_remainder=True)
        testset = testset.batch(batch_size, drop_remainder=True)

        if load_from is None:
            model = pretrainable_component

            build_sample = next(iter(trainset))
            model(build_sample[0])

            optimizer = tf.keras.optimizers.Adam(learning_rate=0.005)
            model.compile(optimizer, loss=PoseEstimationLoss(), metrics=["mse"])
            # model.compile(optimizer, loss="mse", metrics=[rotational_diff_metric, positional_diff_metric])

            def step_decay(epoch, lr):
                drop = 0.5
                epochs_drop = 1.0

                if epoch % epochs_drop == 0 and epoch > 1:
                    return lr * drop
                else:
                    return lr

            # train and save encoder
            model.fit(x=trainset,
                      epochs=epochs,
                      validation_data=valset,
                      callbacks=[
                          tf.keras.callbacks.LearningRateScheduler(step_decay, verbose=1)
                      ],
                      shuffle=True)
            pretrainable_component.save(PRETRAINED_COMPONENTS_PATH + f"/{name}")
        else:
            print("Loading model...")
            model = tf.keras.models.load_model(load_from)
            optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
            model.compile(optimizer, loss="mse", metrics=[rotational_diff_metric, positional_diff_metric])
            print("Model loaded successfully.")

        train_mean = np.mean(list(tfds.as_numpy(trainset.unbatch().take(5000).map(lambda x, y: y))), axis=0)
        test_numpy = np.stack(list(tfds.as_numpy(testset.unbatch().map(lambda x, y: y))))
        print(f"This model achieves {model.evaluate(testset)}")
        print(
            f"A mean model would achieve\n"
            f"\t{np.mean((test_numpy - train_mean) ** 2)} (mse) \n"
            f"\t{tf.reduce_mean(tf.norm(test_numpy[:, :3] - train_mean[:3], axis=-1, ord='euclidean'))} (pos: euclidean)."
        )


def pretrain_on_imagenet(pretrainable_component: tf.keras.Model,
                         epochs: int,
                         name="visual_op",
                         load_from: str = None,
                         batch_size: int = 128):
    """Pretrain a visual component on prediction of cube position."""
    trainset = tfds.load("imagenet_resized/64x64", split="train", shuffle_files=True, as_supervised=True)
    testset = tfds.load("imagenet_resized/64x64", split="validation", shuffle_files=True, as_supervised=True)

    valset, trainset = trainset.take(10000), trainset.skip(10000)

    trainset = trainset.batch(batch_size, drop_remainder=True)
    trainset = trainset.prefetch(AUTOTUNE)

    valset = valset.batch(batch_size, drop_remainder=True)
    valset = valset.prefetch(AUTOTUNE)

    testset = testset.batch(batch_size, drop_remainder=True)
    testset = testset.prefetch(AUTOTUNE)

    if load_from is None:
        model = pretrainable_component
        model(next(iter(trainset))[0])

        # chunk = list(tfds.as_numpy(dataset.take(8000).map(lambda x, y: y)))
        # chunk_mean = np.mean(chunk, axis=0)
        # output_layer = model.get_layer("output")
        # output_weights = output_layer.get_weights()
        # output_weights[1] = chunk_mean
        # output_layer.set_weights(output_weights)

        optimizer = tf.keras.optimizers.Adam(learning_rate=0.003)
        model.compile(optimizer, loss="sparse_categorical_crossentropy", metrics=["sparse_categorical_accuracy",
                                                                                  "sparse_top_k_categorical_accuracy"])

        # train and save encoder
        model.fit(x=trainset,
                  epochs=epochs,
                  validation_data=valset,
                  callbacks=[
                      tf.keras.callbacks.ReduceLROnPlateau(patience=3, factor=0.5, verbose=1)
                  ],
                  shuffle=True)
        pretrainable_component.save(PRETRAINED_COMPONENTS_PATH + f"/{name}")
    else:
        print("Loading model...")
        model = tf.keras.models.load_model(load_from)
        optimizer = tf.keras.optimizers.Adam(learning_rate=0.001)
        model.compile(optimizer, loss="mse", metrics=[rotational_diff_metric, positional_diff_metric])
        print("Model loaded successfully.")

    train_mean = np.mean(list(tfds.as_numpy(trainset.unbatch().take(5000).map(lambda x, y: y))), axis=0)
    test_numpy = np.stack(list(tfds.as_numpy(testset.unbatch().map(lambda x, y: y))))
    print(f"This model achieves {model.evaluate(testset)}")
    print(f"A mean model would achieve {np.mean((test_numpy - train_mean) ** 2)}")


if __name__ == "__main__":
    tf.get_logger().setLevel('INFO')

    # parse commandline arguments
    parser = argparse.ArgumentParser(description="Pretrain a visual component on classification or reconstruction.")

    # general parameters
    parser.add_argument("task", nargs="?", type=str, choices=["classify", "reconstruct", "hand", "object",
                                                              "c", "r", "h", "o", "hp", "op"], default="o")
    parser.add_argument("--name", type=str, default="visual_component",
                        help="Name the pretraining to uniquely identify it.")
    parser.add_argument("--load", type=str, default=None, help=f"load the weights from checkpoint path")
    parser.add_argument("--epochs", type=int, default=25, help=f"number of pretraining epochs")
    parser.add_argument("--batch-size", type=int, default=64, help=f"number of samples per minibatch")

    # read arguments
    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    # parameters
    mode = "pose"
    n_cameras = 3
    n_samples = 83392 // 2
    batch_size = args.batch_size

    visual_component = OpenAIEncoder(shape=(128, 128, 3), name=args.name, n_cameras=n_cameras, mode=mode)
    # visual_component = keras_cortex.cornet.cornet_z.PoseCORNetZ(name=args.name)

    os.makedirs(PRETRAINED_COMPONENTS_PATH, exist_ok=True)

    args.name = args.name + "_" + args.task[0]

    if mode == "classification":
        pretrain_on_imagenet(
            visual_component,
            epochs=args.epochs,
            name=args.name,
            load_from=args.load,
            batch_size=batch_size
        )
    elif mode == "pose":
        pretrain_on_object_pose(
            visual_component,
            epochs=args.epochs,
            batch_size=batch_size,
            n_samples=n_samples,
            n_cameras=n_cameras,
            load_data=True,
            name=args.name,
            load_from=args.load,
        )
    else:
        raise ValueError(f"Unknown mode {mode}.")


"""Custom Exceptions"""

# ENVIRONMENT HANDLING


class UninterpretableObservationSpace(Exception):
    """The environment has an n_steps space that can currently not be handled by our implementation."""
    pass


class IncompatibleModelException(Exception):
    """The model applied to the given environment cannot handle its states."""
    pass


class ComponentError(Exception):
    """Component is incompatible or not loadable."""
    pass



import functools
from typing import List, Callable, Optional
import tensorflow as tf


def hooked_call(inputs, *args, obj: tf.keras.layers.Layer, **kwargs) -> tf.Tensor:
    if obj._before_call is not None:
        obj._before_call(obj, inputs)
    output = obj._old_call(inputs, *args, **kwargs)
    if obj._after_call is not None:
        hook_result = obj._after_call(obj, inputs, output)
        if hook_result is not None:
            output = hook_result
    return output


def register_hook(layers: List[tf.keras.layers.Layer],
                  before_call: Callable[[tf.keras.layers.Layer, tf.Tensor], None] = None,
                  after_call: Callable[[tf.keras.layers.Layer, tf.Tensor, tf.Tensor], Optional[tf.Tensor]] = None):
    for layer in layers:
        layer._before_call = before_call
        layer._after_call = after_call
        layer._old_call = layer.call
        layer.call = functools.partial(hooked_call, obj=layer)


# todo this could be done through a context
def clear_hooks(module: tf.keras.Model):
    """Remove all hooks from a model."""

    for layer in module.submodules:
        if not hasattr(layer, "_old_call"):
            continue

        del layer._before_call
        del layer._after_call
        layer.call = layer._old_call
        del layer._old_call


from typing import List, Dict

import bokeh.layouts
from bokeh import embed
from bokeh.plotting import figure

from angorapy.utilities.monitor.plotting_base import palette, plot_styling, style_plot
import numpy as np
import pandas as pd


def plot_episode_box_plots(rewards: List[float], lengths: List[float]):
    """Boxplot of the rewards and lengths given."""
    if len(rewards) < 1:
        return "", ""

    cats = ["reward"]
    rewards = np.array(rewards, dtype=np.float)

    # find the quartiles and IQR for each category
    q1 = np.quantile(rewards, q=0.25, axis=-1)
    q2 = np.quantile(rewards, q=0.5, axis=-1)
    q3 = np.quantile(rewards, q=0.75, axis=-1)
    iqr = q3 - q1
    upper = q3 + 1.5 * iqr
    lower = q1 - 1.5 * iqr

    # find the outliers for each category
    # def outliers(group):
    #     cat = group.name
    #     return group[(group.score > upper.loc[cat]['score']) | (group.score < lower.loc[cat]['score'])]['score']
    #
    # out = groups.apply(outliers).dropna()

    # prepare outlier rewards for plotting, we need coordinates for every outlier.
    # if not out.empty:
    #     outx = list(out.index.get_level_values(0))
    #     outy = list(out.values)

    p = figure(title="Preprocessor Running Mean",
               x_axis_label='Cycle',
               y_axis_label='Running Mean',
               **plot_styling,
               x_range=cats
    )

    # if no outliers, shrink lengths of stems to be no longer than the minimums or maximums
    # qmin = groups.quantile(q=0.00)
    # qmax = groups.quantile(q=1.00)
    # upper.score = [min([x, y]) for (x, y) in zip(list(qmax.loc[:, 'score']), upper.score)]
    # lower.score = [max([x, y]) for (x, y) in zip(list(qmin.loc[:, 'score']), lower.score)]

    # stems
    p.segment(cats, upper, cats, q3, line_color="black")
    p.segment(cats, lower, cats, q1, line_color="black")

    # boxes
    p.vbar(cats, 0.7, q2, q3, fill_color="#E08E79", line_color="black")
    p.vbar(cats, 0.7, q1, q2, fill_color="#3B8686", line_color="black")

    # whiskers (almost-0 height rects simpler than segments)
    p.rect(cats, lower, 0.2, 0.01, line_color="black")
    p.rect(cats, upper, 0.2, 0.01, line_color="black")

    # outliers
    # if not out.empty:
    #     p.circle(outx, outy, size=6, color="#F38630", fill_alpha=0.6)

    subsamplesize = 100
    if len(rewards) > subsamplesize:
        rewards = np.random.choice(rewards, subsamplesize)

    p.circle(cats * len(rewards), rewards, size=6, color="#F38630", fill_alpha=0.6)

    p.xgrid.grid_line_color = None
    p.ygrid.grid_line_color = "white"
    p.grid.grid_line_width = 2
    p.xaxis.major_label_text_font_size = "16px"

    p.legend.location = "bottom_right"
    style_plot(p)

    return p


def plot_per_receptor_mean(per_receptor_mean: Dict[str, List]):
    plots = []
    for i, (sense, means) in enumerate(per_receptor_mean.items()):
        p = figure(title=f"{sense.capitalize()} Distribution",
                   x_axis_label='Receptor',
                   y_axis_label='Mean',
                   **plot_styling,
                   )

        p.vbar(range(len(means)), width=1, top=means, legend_label=sense, color=palette[i])
        p.legend.location = "bottom_right"
        style_plot(p)

        plots.append(p)

    p = bokeh.layouts.column(plots)
    p.sizing_mode = "stretch_width"
    return p



from angorapy.analysis.investigators import Predictability

import tensorflow as tf

from angorapy.common.policies import BasePolicyDistribution
import tensorflow_probability as tfp


class Correlation(Predictability):

    def __init__(self, network: tf.keras.Model, distribution: BasePolicyDistribution):
        super().__init__(network, distribution)

    def fit(self, source_layer: str, target_information: str):
        """Measure the correlation of target_information with source_layer's activation."""
        assert self.prepared, "Need to prepare before investigating."
        assert source_layer in self.list_layer_names(only_para_layers=False) + ["noise"]
        assert target_information in self._data.as_numpy_iterator().__next__().keys()

        all_cross_correlations = []
        for source, target in zip(list(self._data.map(lambda x: x[source_layer]).as_numpy_iterator()),
                                  list(self._data.map(lambda x: x[target_information]).as_numpy_iterator())):

            source = source[..., tf.newaxis]
            target = target[..., tf.newaxis, tf.newaxis]

            cross_correlation = tf.reduce_mean(tf.nn.convolution(source, target))
            all_cross_correlations.append(cross_correlation)

        mean_correlation = tf.reduce_mean(all_cross_correlations)

        print(f"{source_layer} -> {target_information}: {mean_correlation}")



import itertools
import json
import os

import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.stats.api as sms
from matplotlib import pyplot as plt

from angorapy.common.const import PATH_TO_EXPERIMENTS, QUALITATIVE_COLOR_PALETTE

# experiment_ids = [['1674343261520188', '1674308148646663', '1674074113967956', '1674074113731734', '1673350499432390']]  # only best setting
experiment_ids = [['1674343261520188', '1674308148646663', '1674074113967956', '1673350499432390', '1671749718306373',
                   '1674975602446141', '1674975602294059', '1674074113731734', '1673786170549564'],
                  ['1674975602294059']]  # only best setting
# experiment_ids = [['1674074113967956', '1673350499432390', '1674074113731734']]  # only best setting
names = [f"mean (n={len(experiment_ids[0])})", "best"]

# experiment_ids = [['1673786170549564'], ['1673350499432390']]  # shared vs unshared
# experiment_ids = [['1674975602294059', '1674975602446141', '1671749718306373'],
#                   ['1674343261520188', '1674308148646663', '1674074113967956', '1674074113731734', '1673350499432390']]  # symmetric vs asymmetric
# names = ["symmetric", "asymmetric"]

# experiment_ids = [['1673350499432390']]
# names = ["asymmetric"]

# experiment_ids = [['1675028736765791', '1674983643322591'],
#                   ['1674985163288377', '1674983177286330'],
#                   ['1674343261520188', '1674308148646663', '1674074113967956', '1673350499432390']]  # compare distributions
# names = ["beta", "gaussian", "multicategorical"]
# names = ["beta", "multicategorical"]
reward_developments = {}
reward_bands = {}

cosucc_developments = {}
cosucc_bands = {}

evaluation_rewards = {}
evaluation_reward_histograms = {}

for i, group in enumerate(experiment_ids):
    exp_name = names[i]
    reward_developments[exp_name] = []
    cosucc_developments[exp_name] = []
    evaluation_rewards[exp_name] = []
    evaluation_reward_histograms[exp_name] = []

    for id in group:
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "meta.json")) as f:
            meta = json.load(f)
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "progress.json")) as f:
            progress = json.load(f)
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "statistics.json")) as f:
            statistics = json.load(f)

        try:
            with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "evaluation.json")) as f:
                evaluation = json.load(f)

            evaluation_rewards[exp_name] += evaluation["auxiliary_performances"]["consecutive_goals_reached"]
            evaluation_reward_histograms[exp_name].append(
                np.histogram(evaluation["auxiliary_performances"]["consecutive_goals_reached"], bins=50))
        except:
            print(f"no evaluation for {id}")

        reward_developments[exp_name].append(progress["rewards"]["mean"])
        cosucc_developments[exp_name].append(statistics["auxiliary_performances"]["consecutive_goals_reached"]["mean"])

    try:
        evaluation_reward_histograms[exp_name] = np.mean([c for c, b in evaluation_reward_histograms[exp_name]],
                                                         axis=0), \
            evaluation_reward_histograms[exp_name][0][1]
    except:
        pass

    max_length = min(1700, max(len(arr) for arr in cosucc_developments[exp_name]))

    reward_developments[exp_name] = np.array(
        [trace[:max_length] for trace in reward_developments[exp_name]])
    reward_developments[exp_name] = np.array(
        [np.pad(arr, (0, max_length - len(arr)), mode='maximum') for arr in reward_developments[exp_name]])
    cosucc_developments[exp_name] = np.array(
        [trace[:max_length] for trace in cosucc_developments[exp_name]])
    cosucc_developments[exp_name] = np.array(
        [np.pad(arr, (0, max_length - len(arr)), mode='maximum') for arr in cosucc_developments[exp_name]])

    rd_mean = np.mean(reward_developments[exp_name], axis=0)
    cd_mean = np.mean(cosucc_developments[exp_name], axis=0)
    rd_std = np.std(reward_developments[exp_name], axis=0)
    cd_std = np.std(cosucc_developments[exp_name], axis=0)

    descm = sms.DescrStatsW(reward_developments[exp_name])
    rd_mean = np.mean(reward_developments[exp_name], axis=0)
    reward_bands[exp_name] = descm.tconfint_mean()

    descm = sms.DescrStatsW(cosucc_developments[exp_name])
    cd_mean = np.mean(cosucc_developments[exp_name], axis=0)
    cosucc_bands[exp_name] = descm.tconfint_mean()

if len(names) <= 2:
    axes = [
        plt.subplot2grid((1, 5), (0, 0), colspan=2),
        plt.subplot2grid((1, 5), (0, 2), colspan=2),
        plt.subplot2grid((1, 5), (0, 4), colspan=1),
    ]
else:
    axes = [
        plt.subplot2grid((1, 3), (0, 0), colspan=1),
        plt.subplot2grid((1, 3), (0, 1), colspan=1),
        plt.subplot2grid((1, 3), (0, 2), colspan=1),
    ]

x_max = len(list(reward_developments.items())[0][1][0])
y_max_x0 = -np.inf
for i, (name, rewards) in enumerate(reward_developments.items()):
    x_max = max(x_max, len(rewards[0]))
    axes[0].plot(np.mean(rewards, axis=0), label=name, color=QUALITATIVE_COLOR_PALETTE[i])
    axes[0].fill_between(range(len(reward_bands[name][0])), reward_bands[name][0], reward_bands[name][1], alpha=0.2)
    y_max_x0 = max(y_max_x0, np.max(reward_bands[name][1]))

    axes[1].plot(np.mean(cosucc_developments[name], axis=0), label=name, color=QUALITATIVE_COLOR_PALETTE[i])
    axes[1].fill_between(range(len(cosucc_bands[name][0])), cosucc_bands[name][0], cosucc_bands[name][1], alpha=0.2)

    # # plot horizontal line indicating max performance with circles as markers
    # axes[1].axhline(y=np.max(mean_line), color=QUALITATIVE_COLOR_PALETTE[i], linestyle="--", linewidth=1)
    #
    # # plot horizontal line indicating max performance of best agent in group
    # axes[1].axhline(y=np.max(np.max(cosucc_developments[name], axis=0)), linestyle=":", color=QUALITATIVE_COLOR_PALETTE[i], linewidth=0.5)

    # axes[2].hist(
    #     evaluation_reward_histograms[name][1][:-1],
    #     evaluation_reward_histograms[name][1],
    #     weights=evaluation_reward_histograms[name][0],
    #     edgecolor="white", linewidth=0.5, alpha=0.8
    # )
    # axes[2].set_xlabel("Consecutive Goals Reached")
    # axes[2].set_ylabel("Number of Episodes")

df = pd.DataFrame(
    {"group": itertools.chain(*[[name] * len(dp) for name, dp in evaluation_rewards.items()]),
     "Consecutive Goals Reached": np.concatenate([dp for name, dp in evaluation_rewards.items()])}
)

sns.boxplot(data=df, x="group", y="Consecutive Goals Reached", notch=True, medianprops={"color": "red"},
            flierprops={"marker": "x"},
            fliersize=1, ax=axes[2], palette={name: QUALITATIVE_COLOR_PALETTE[i] for i, name in enumerate(names)},
            showmeans=True, meanprops={"marker": "o", "markerfacecolor": "white", "markeredgecolor": "black"})
axes[2].set_xlabel("")
if len(names) == 1:
    axes[2].set_xticks([])

modes = df.groupby("group").aggregate(pd.Series.mode)
# scatter modes as unfilled triangles on top of boxplot
axes[2].scatter(modes.index, modes["Consecutive Goals Reached"],
                marker="^", linewidth=1, facecolors=None,
                zorder=10, edgecolors="red")

axes[0].set_xlabel("Cycle")
axes[0].set_ylabel("Avg. Episode Return")

if len(names) > 1:
    axes[0].legend()

axes[1].set_xlabel("Cycle")
axes[1].set_ylabel("Avg. Consecutive Goals Reached")

if len(names) > 1:
    axes[1].legend()

axes[0].set_xlim(0, x_max)
axes[0].set_ylim(top=y_max_x0 * 1.1)
axes[1].set_xlim(0, x_max)
axes[1].set_ylim(0, 50)
axes[2].set_ylim(0, 52)

plt.gcf().set_size_inches(12, 4)
plt.subplots_adjust(wspace=0.5, right=0.995, left=0.05, top=0.98)
# plt.show()
plt.savefig("../../../docs/figures/manipulate-learning-curves.pdf", format="pdf")
